# -*- coding: utf-8 -*-

#  This software and supporting documentation are distributed by
#      Institut Federatif de Recherche 49
#      CEA/NeuroSpin, Batiment 145,
#      91191 Gif-sur-Yvette cedex
#      France
#
# This software is governed by the CeCILL-B license under
# French law and abiding by the rules of distribution of free software.
# You can  use, modify and/or redistribute the software under the
# terms of the CeCILL-B license as circulated by CEA, CNRS
# and INRIA at the following URL "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL-B license and that you accept its terms.

'''
This module contains functions and classes related to sqlite databases.

* author: Yann Cointepas
* organization: NeuroSpin
* license: `CeCILL B <http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html>`_
'''
from __future__ import absolute_import
from __future__ import print_function
__docformat__ = "restructuredtext en"


import sys
import threading
import sqlite3

#------------------------------------------------------------------------------


class ThreadSafeSQLiteConnection(object):

    '''
    Python wrapping of SQLite do not allow sharing of database connection between
    threads. This class allows to automatically create a connection for each
    thread.
    '''
    _currentId = 0
    _classLock = threading.RLock()

    def __init__(self, *args, **kwargs):
        '''
        A ThreadSafeSQLiteConnection is created with the parameters that
        would be used for a call to sqlite3.connect() in a single thread
        system. These parameters are stored to allow to create a separate
        SQLite connection for any thread with get_connection().
        '''
        super(ThreadSafeSQLiteConnection, self).__init__()
        self.__args = args
        self.__kwargs = kwargs
        self._instanceLock = threading.RLock()
        self.connections = {}
        self._classLock.acquire()
        try:
            self._id = ThreadSafeSQLiteConnection._currentId
            ThreadSafeSQLiteConnection._currentId += 1
        finally:
            self._classLock.release()

    def __del__(self):
        if threading is None:
            # The interpretor is exiting and we cannot access threading
            # module. We cannot do anything.
            return
        if self.__args is not None:
            sqliteFile = self.__args[0]
            try:
                self.close()
                for thread in self.connections.keys():
                    connection, connectionClosed = self.connections[thread]
                    if connection is not None:
                        currentThread = threading.current_thread().getName()
                        print('WARNING: internal error: an sqlite connection on',
                              repr(sqliteFile), 'is opened for thread', thread,
                              'but the corresponding ThreadSafeSQLiteConnection instance (number '
                              + str(self._id)
                              + ') is being deleted in thread', currentThread
                              + '. Method currentThreadCleanup() should have been called from',
                              thread, 'to supress this warning.', file=sys.stderr)
            except ImportError:
                # python is shutting down
                pass

    def get_connection(self):
        '''
        Returns a SQLite connection (i.e. the result of sqlite3.connect)
        for the current thread. If it does not already exists, it is created
        and stored for the current thread. In order to destroy this connection
        it is necessary to call self.delete_connection() from the current
        thread. If a ThreadSafeSQLiteConnection is destroyed in a thread,
        all connections opened in other threads must have been deleted.
        '''
        if self.__args is None:
            raise RuntimeError(
                'Attempt to access to a closed ThreadSafeSQLiteConnection')
        currentThread = threading.current_thread().getName()
        # print('!ThreadSafeSQLiteConnection:' + currentThread + '!')
        # _getConnection( id =', self._id, ')', self.__args
        self._instanceLock.acquire()
        try:
            # currentThreadConnections = self.connections.setdefault( currentThread, {} )
            # print('!ThreadSafeSQLiteConnection:' + currentThread + '!')
            # currentThreadConnections =', currentThreadConnections
            connection, connectionClosed = self.connections.get(
                currentThread, (None, True))
            if connectionClosed:
                if connection is not None:
                    connection.close()
                connection = sqlite3.connect(*self.__args, **self.__kwargs)
                # print('!ThreadSafeSQLiteConnection:' + currentThread + '!'
                # opened', connection)
                self.connections[currentThread] = (connection, False)
        finally:
            self._instanceLock.release()
        return connection

    def delete_connection(self):
        '''
        Delete the connection previously created for the current thread with
        get_connection()
        '''
        if threading.current_thread is None:
            # exiting, threading attributes have become None
            return
        currentThread = threading.current_thread().getName()
        self._instanceLock.acquire()
        try:
            connection, connectionClosed = self.connections.pop(
                currentThread, (None, True))
        finally:
            self._instanceLock.release()
        if connection is not None:
            connection.close()

    def close(self):
        '''
        After this call no more connection can be opened
        '''
        if self.__args is not None:
            self.closeSqliteConnections()
            self.__args = None
            self.__kwargs = None

    def close_connections(self):
        '''
        Mark the opened connection for all the threads as closed.
        Subsequent calls to get_connection() will have to
        recreate the sqlite connection with sqlite3.connect().
        This method does not delete the connection of the current
        thread.
        '''
        if self.__args is not None:
            if threading.current_thread is None:
                # exiting, threading attributes have become None
                return
            self.currentThreadCleanup()
            currentThread = threading.current_thread().getName()
            self._instanceLock.acquire()
            try:
                for thread in self.connections.keys():
                    connection, connectionClosed = self.connections[thread]
                    self.connections[thread] = (connection, True)
            finally:
                self._instanceLock.release()

    # For backward compatibility
    _getConnection = get_connection
    currentThreadCleanup = delete_connection
    closeSqliteConnections = close_connections



# -*- coding: utf-8 -*-
#  This software and supporting documentation are distributed by
#      Institut Federatif de Recherche 49
#      CEA/NeuroSpin, Batiment 145,
#      91191 Gif-sur-Yvette cedex
#      France
#
# This software is governed by the CeCILL license version 2 under
# French law and abiding by the rules of distribution of free software.
# You can  use, modify and/or redistribute the software under the
# terms of the CeCILL license version 2 as circulated by CEA, CNRS
# and INRIA at the following URL "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license version 2 and that you accept its terms.

"""
Utils for socket communication
"""
from __future__ import print_function

from __future__ import absolute_import
import threading
import socket
import errno
import time
import sys

import six
import six.moves.queue as queue

from soma.qt_gui.qt_backend.QtCore import QObject, QSocketNotifier


class Socket(QObject):

    """
    Opens a connection to a socket server and provides methods to read from and write to socket streams.
    To handle specific message format, redefine readMessage method. By default it reads a line on the socket.
    To process messages, redefine processMessage method. By default it prints the message to standard output.

    Attributes
    ----------
    dest: string
        socket server machine
    port: int
        port that the socket server listens
    socket: socket.socket
        python socket object
    socketnotifier: qt.QSocketNotifier
        the notifier sends a signal when there's something to read on the
        socket.
    readLock: threading.Lock
        lock to prevent threads from reading at the same time on the socket
    writeLock: threading.Lock
        lock to prevent threads from writing at the same time on the socket
    lock: threading.RLock
        lock to prevent concurrent access on object data because it can be used
        by multiple threads. At least principal thread and reading messages
        thread when it not possible to use a QSocketNotifier.
    initialized: bool
        indicates if the connection is correctly opened
    notifyenabled: bool
        indicates if message received on the socket must be processed.

    loopRetry: int
        class attribute: max number of connection tries
    defaultPort: int
        class attribute: default port for socket server
    """

    loopRetry = 60      # Retry to connect 60 times (1 minute)
    defaultPort = 50007

    def __init__(self, host, port=None):
        """
        Parameters
        ----------
        host: string
            socket server machine (localhost if it is current machine)
        port: int
            port that the socket server listens
        """
        super(Socket, self).__init__()
        self.dest = host
        self.port = port
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socketnotifier = None
        # We are in a multi threaded environnement, so only one thread can
        # use the socket at the same time
        self.readLock = threading.Lock()
        self.writeLock = threading.Lock()
        self.lock = threading.RLock()
        self.initialized = 0
        self.notifyenabled = 1

    def initialize(self, port=None):
        """
        Connects the socket to the server and sets the frame to read data from the socket.
        Two methods for reading:

        - reading thread
        - QSocketNotifier (not possible on windows platform)

        Returns
        -------
        port: int
            port that the socket server listens
        """
        if self.initialized:
            return

        if port is not None:
            self.port = port
        # We have to retry connecting because it can take time for the socket
        # server to start
        i = 0
        while i < Socket.loopRetry:
            # using socket.htons to convert port number is useless
            # because Python take care of this
            res = self.socket.connect_ex((self.dest, self.port))
            if res != 0:
                # retry to connect
                time.sleep(1)
                self.socket.close()
                self.socket = socket.socket(socket.AF_INET,
                                            socket.SOCK_STREAM)
                i = i + 1
            else:
                # Connection successful, exit from the while loop
                break

        if findPlatform() == 'windows':
            # on windows, QSocketNotifier seems to make the socket fail
            self.usethread = 1
        else:
            self.usethread = 1  # 0 # QSocketNotifier seems to create problems when it is not used in qt main thread (application freeze)
        # sets the frame to read from the socket
        if self.usethread:
            # reading thread
            self.readthread = threading.Thread(target=self.readForever)
            self.readthread.setDaemon(True)  # don't block when exiting program
            self._messages = queue.Queue(500)
            self.socket.setblocking(0)
            self.initialized = 1
            self.readthread.start()
        else:
            # QSocketNotifier that signals that data is enable for reading
            self.socketnotifier = QSocketNotifier(
                self.socket.fileno(), QSocketNotifier.Read)  # , self )
            self.socketnotifier.activated.connect(self.messageHandler)
            self.socketnotifier.setEnabled(True)
            self.initialized = 1

    def messageHandler(self):
        """
        This method is called when a message is received on the socket.
        Gets and processes the message.
        """
        excep = None
        msg = None
        try:
            msg = self.getMessage()
        except Exception as e:
            excep = e
        self.processMessage(msg, excep)

    def getMessage(self, timeout=30):
        '''
        Called by message handler to gets the last received message.
        If socket communication is managed by a QSocketNotifier, the message is not yet read, so this method read it.
        If reading is in another process, the message is already read and it is in messages queue.
        '''
        if not self.initialized:
            return '', ''
        if self.usethread:
        # block until a message arrives in queue
            self.lock.acquire()
            try:
                try:
                    msg = self._messages.get(True, timeout)
                except queue.Empty as e:
                    raise IOError(
                        errno.ETIMEDOUT,  'socket communication timed out')
            finally:
                self.lock.release()
        else:
            try:
                msg = self.readMessage(timeout)
            except IOError as e:
                if e.errno == errno.EPIPE:
                    # The socket has been closed
                    # Return to avoid infinite loop
                    self.close()
                raise
        return msg

    def processMessage(self, msg, excep):
        """
        Processes a message received on the socket.
        This method only print the message. To do some specific treatment, subclass Socket and redefine this method.
        """
        print("message received :", msg, excep)

    def send(self, msg):
        """
        Sends data to the connected socket.
        @type msg: string
        @param msg: the message to send.
        """
        if not self.initialized:
            return
        totalsent = 0
        msglen = len(msg)
        # Send the message atomically (send is not thread-safe)
        close = False
        self.writeLock.acquire()
        # sendall() only appeared in python 2.1.12
        # self.socket.sendall( msg )
        if hasattr(msg, 'encode'):
            # encode to bytes (python3)
            msg = msg.encode()
        n = 0
        try:
            while n < msglen:
                try:
                    n += self.socket.send(msg[n:])
                except socket.error as e:
                    if e.errno == errno.EWOULDBLOCK:
                        time.sleep(0.02)
                    else:
                        close = True
                        break
        finally:
            self.writeLock.release()
        if close:
            self.close()

    def readForever(self):
        """
        Reading loop to handle reading messages from the socket.
        Used only if QSocketNotifier cannot be used.
        """
        while self.initialized:
            try:
                msg = self.readMessage()
                self.lock.acquire()
                try:
                    self._messages.put(msg)
                finally:
                    self.lock.release()
                if self.notifyenabled:
                    self.messageHandler()
            except IOError as e:
                if e.errno != errno.ETIMEDOUT:
                    self.close()
                    # raise
            if not self.initialized:
                break

    def readLine(self, timeout):
        """
        Reads a line of data from the socket (a string followed by ``'\\n'``).
        self.readLock must be acquired before calling this method.
        If data cannot be read before timeout (in seconds), an IOError exception is raised.

        Returns
        -------
        timeout: int
            max time to wait before reading the message.
        """
        msg = b''
        char = b''
        waitedTime = 0
        # Receive the message atomically
        while char != b'\n':
            if char != b'':
                msg += char
            try:
                char = self.socket.recv(1)
                waitedTime = 0
                if char == b'\0' or char == b'':
                    e = IOError(
                        errno.EPIPE, 'socket communication interrupted')
                    raise e
            except socket.error as e:
                if e.errno == errno.EWOULDBLOCK:
                    char = b''
                    time.sleep(0.02)
                    waitedTime += 0.02
                    if waitedTime >= timeout:
                        raise IOError(
                            errno.ETIMEDOUT, 'socket communication timed out')
                else:
                    raise IOError(
                        errno.EPIPE, 'socket communication interrupted')
        return six.ensure_str(msg)

    def readMessage(self, timeout=30):
        """
        Reads a message from the socket. This method only gets the readlock and reads a line.
        To read specific message formats, subclass Socket and redefine this method.

        Parameters
        ----------
        timeout: int
            max time to wait before reading the message.

        Returns
        -------
        message: string
            the message received from the socket
        """
        self.readLock.acquire()
        try:
            msg = self.readLine(timeout)
        finally:
            self.readLock.release()
        return msg

    def findFreePort(self):
        """
        Try to find the first unused port.
        @rtype: int
        @return: numero of a free port
        """
        if self.port is not None:
            startport = self.port
        else:
            startport = Socket.defaultPort
        res = 0
        tmpsocket = socket.socket(socket.AF_INET,
                                  socket.SOCK_STREAM)
        while res == 0:
            res = tmpsocket.connect_ex((self.dest, startport))
            tmpsocket.close()
            if res != 0:
                return startport
            tmpsocket = socket.socket(socket.AF_INET,
                                      socket.SOCK_STREAM)
            startport += 1

    def disableHandler(self):
        """Disables reading from the socket."""
        self.notifyenabled = 0
        if self.socketnotifier is not None:
            self.socketnotifier.setEnabled(0)

    def enableHandler(self):
        """Enables reading from the socket."""
        self.notifyenabled = 1
        if self.socketnotifier is not None:
            self.socketnotifier.setEnabled(1)

    def __del__(self):
        """
        Closes the connection when the object is deleted
        """
        self.close()

    def close(self):
        """
        Closes the connection.
        """
        if not self.initialized:
            return
        self.initialized = 0
        self.disableHandler()
        self.socketnotifier = None
        self.socket.close()
        if self.usethread and self.readthread != threading.current_thread():
            self.readthread.join()


def findPlatform():
    """
    Identify platform, possible return values are :
      - 'windows': Windows
      - 'linux': Linux
      - 'sunos': SunOS (Solaris)
      - 'darwin': Darwin (MacOS X)
      - None: unknown
    """
    if sys.platform[:3] == "win":
        return 'windows'
    else:
        if sys.platform.find('linux') != -1:
            return 'linux'
        if sys.platform.find('sunos') != -1:
            return 'sunos'
        if sys.platform.find('darwin') != -1:
            return 'darwin'
        if sys.platform.find('irix') != -1:
            return 'irix'
        return sys.platform



# -*- coding: utf-8 -*-

from __future__ import print_function

from __future__ import absolute_import
import unittest
import os
import sys
from soma import test_utils
from soma.test_utils import base_classes
from soma.test_utils import test_notebook

class TestTestUtils(unittest.TestCase):

    def test_test_classes(self):
        pass

    if test_notebook.main_jupyter is not None:
        def test_test_notebook(self):
            pass


def test():
    suite = unittest.TestLoader().loadTestsFromTestCase(TestTestUtils)
    runtime = unittest.TextTestRunner(verbosity=2).run(suite)
    return runtime.wasSuccessful()


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-

#  This software and supporting documentation are distributed by
#      Institut Federatif de Recherche 49
#      CEA/NeuroSpin, Batiment 145,
#      91191 Gif-sur-Yvette cedex
#      France
#
# This software is governed by the CeCILL-B license under
# French law and abiding by the rules of distribution of free software.
# You can  use, modify and/or redistribute the software under the
# terms of the CeCILL-B license as circulated by CEA, CNRS
# and INRIA at the following URL "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL-B license and that you accept its terms.


'''
soma module documentation is not finished yet

Examples
========

Insert examples here

Module content
==============

The content of soma module is composed of many submodules. However,
all important items can be imported from soma.api. For example, if
one wants to use the Application class defined in the module
soma.application, he just have to use the following import
statement::

  from soma.api import Application

Main classes
------------
- :class:`Application <soma.application.Application>`: an Application instance
  contains all information that is shared across all modules of an application.
- :class:`Controller <soma.controller.controller.Controller>`

Advanced classes
----------------
- :class:`Singleton <soma.singleton.Singleton>`: A class deriving from
  Singleton can have only one instance.

Information
-----------
- Yann Cointepas
- NeuroSpin
- License: CeCILL B (http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html)
'''


from __future__ import absolute_import
from soma.singleton import Singleton
from soma.application import Application
from soma.controller import Controller



# -*- coding: utf-8 -*-
"""
This module contains some classes to manage python tests.

Our tests may need to create a reference for later comparison. As we want to be
able to use the same code, there are several "modes":
  - in the ref mode, the code is supposed to create the reference data
  - in the run mode, the code is supposed to create data and perform the
    comparison
The different modes are invoked by bv_maker. This works by using some CLI
arguments and by setting environment variables.

To do so, we need our own test loader that will configure the test cases. This
is done by modifying attributes at the class level (so that setUpClass can
know in which mode it is).
"""

from __future__ import absolute_import
import os
import argparse

import unittest

ref_mode = 'ref'
run_mode = 'run'
test_modes = [run_mode, ref_mode]
default_mode = run_mode


class SomaTestLoader(unittest.TestLoader):
    """Base class for test loader that allows to pass keyword arguments to the
       test case class and use the environment to set the location of reference
       files. Inspired from http://stackoverflow.com/questions/11380413/ (but
       here we modify the test case classes themselves).
       The ArgumentParser is defined in the __init__ method. While it could be
       a class atrtibute, this makes easier to modify it in subclasses.
       The static method `default_parser` is used to create a new instance of
       the basic parser.
    """

    @staticmethod
    def default_parser(description):
        parser = argparse.ArgumentParser(
            description=description,
            epilog=("Note that the options are usually passed by make via "
                    "bv_maker.")
        )
        parser.add_argument(
            '--test_mode', choices=test_modes,
            default=default_mode,
            help=('Mode to use (\'run\' for normal tests, '
                  '\'ref\' for generating the reference files).'))
        return parser

    def __init__(self):
        self.parser = SomaTestLoader.default_parser("Soma test program.")

    def parse_args_and_env(self, argv):
        args = vars(self.parser.parse_args(argv))
        args['base_ref_data_dir'] = os.environ.get(
            'BRAINVISA_TEST_REF_DATA_DIR'
        )
        args['base_run_data_dir'] = os.environ.get(
            'BRAINVISA_TEST_RUN_DATA_DIR'
        )
        if args['test_mode'] == run_mode and not args['base_run_data_dir']:
            msg = ("BRAINVISA_TEST_RUN_DATA_DIR must be set in environment "
                   "when using 'run' mode")
            raise ValueError(msg)
        return args

    def loadTestsFromTestCase(self, testCaseClass, argv=None):
        """Return a suite of all tests cases contained in
           testCaseClass."""
        if issubclass(testCaseClass, unittest.TestSuite):
            raise TypeError("Test cases should not be derived from "
                            "TestSuite. Maybe you meant to derive from"
                            " TestCase?")
        testCaseNames = self.getTestCaseNames(testCaseClass)
        if not testCaseNames and hasattr(testCaseClass, 'runTest'):
            testCaseNames = ['runTest']

        args = self.parse_args_and_env(argv)

        # Modification here: configure testCaseClass.
        test_cases = []
        for test_case_name in testCaseNames:
            for k, v in args.items():
                setattr(testCaseClass, k, v)
            test_cases.append(
                testCaseClass(test_case_name)
            )
        loaded_suite = self.suiteClass(test_cases)

        return loaded_suite


class BaseSomaTestCase(unittest.TestCase):
    """
    Base class for test cases that honor the options to create reference files.
    The base location for reference data and run data are set in the class by
    the loader. Subclasses can define private_dir to have a specific
    sub-directory inside those directories.
    This class don't define any setUp method so direct subclasses of this class
    can implement any scenario.
    """

    # Declare base class attributes
    test_mode = None
    base_ref_data_dir = None
    base_run_data_dir = None

    # Subclasses should define this variable to create a private folder (if
    # it's None, the base directory for run or ref will be used).
    private_dir = None

    # Computed in property
    _private_ref_data_dir = None
    _private_run_data_dir = None

    def __init__(self, testName):
        super(BaseSomaTestCase, self).__init__(testName)
        if self.test_mode == run_mode and not self.base_run_data_dir:
            msg_fmt = \
                "base_run_data_dir must be provided when using '%s' mode"
            msg = msg_fmt % run_mode
            raise ValueError(msg)

    @classmethod
    def private_ref_data_dir(cls):
        if cls.private_dir:
            if not cls._private_ref_data_dir:
                cls._private_ref_data_dir = os.path.join(
                    cls.base_ref_data_dir, cls.private_dir
                )
            return cls._private_ref_data_dir
        else:
            return cls.base_ref_data_dir

    @classmethod
    def private_run_data_dir(cls):
        # base_run_data_dir is None in run mode
        if cls.test_mode == ref_mode:
            return None
        if cls.private_dir:
            if not cls._private_run_data_dir:
                cls._private_run_data_dir = os.path.join(
                    cls.base_run_data_dir, cls.private_dir
                )
            return cls._private_run_data_dir
        else:
            return cls.base_run_data_dir


class SomaTestCaseWithoutRefFiles(BaseSomaTestCase):
    """
    Special test case that don't need reference file (i.e. should not be run in
    ref mode).
    """

    def __init__(self, testName):
        super(SomaTestCaseWithoutRefFiles, self).__init__(testName)
        if self.test_mode == ref_mode:
            msg_fmt = "Test %s should not be run in '%s' mode"
            msg = msg_fmt % (self.__class__.__name__, ref_mode)
            raise EnvironmentError(msg)


# Should we skip all tests in ref mode (all computation in setUp_ref_mode?)
class SomaTestCase(BaseSomaTestCase):
    """
    Base class for tests that need simple customization for ref and run modes.
    Subclasses can redefine `setUpClass_ref_mode`, `setUpClass_run_mode`,
    `setUp_ref_mode` and `setUp_run_mode` (and corresponding methods for tear
    down).

    Note that the test will be skipped if ref data dir is not defined.
    """

    @classmethod
    def setUpClass_ref_mode(cls):
        """
        This method is called by setUpClass in ref mode.
        """
        pass

    @classmethod
    def setUpClass_run_mode(cls):
        """
        This method is called by setUpClass in run mode.
        """
        pass

    @classmethod
    def setUpClass(cls):
        if not cls.base_ref_data_dir:
            msg_fmt = \
                "base_ref_data_dir must be provided for test %s"
            msg = msg_fmt % cls.__name__
            raise EnvironmentError(msg)
        if cls.test_mode == ref_mode:
            try:
                os.makedirs(cls.private_ref_data_dir())
            except OSError:
                pass
            cls.setUpClass_ref_mode()
        else:
            try:
                os.makedirs(cls.private_run_data_dir())
            except OSError:
                pass
            cls.setUpClass_run_mode()

    @classmethod
    def tearDownClass_ref_mode(cls):
        """
        This method is called by tearDownClass in ref mode.
        """
        pass

    @classmethod
    def tearDownClass_run_mode(cls):
        """
        This method is called by tearDownClass in run mode.
        """
        pass

    @classmethod
    def tearDownClass(cls):
        if cls.test_mode == ref_mode:
            cls.tearDownClass_ref_mode()
        else:
            cls.tearDownClass_run_mode()

    def setUp_ref_mode(self):
        """
        This method is called by setUp in ref mode.
        """
        pass

    def setUp_run_mode(self):
        """
        This method is called by setUp in run mode.
        """
        pass

    def setUp(self):
        if self.test_mode == ref_mode:
            self.setUp_ref_mode()
        else:
            self.setUp_run_mode()

    def tearDown_ref_mode(self):
        """
        This method is called by tearDown in ref mode.
        """
        pass

    def tearDown_run_mode(self):
        """
        This method is called by tearDown in run mode.
        """
        pass

    def tearDown(self):
        if self.test_mode == ref_mode:
            self.tearDown_ref_mode()
        else:
            self.tearDown_run_mode()



# -*- coding: utf-8 -*-
#
# SOMA - Copyright (C) CEA, 2015
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
#

# System import
from __future__ import print_function

from __future__ import absolute_import
import os
import logging
import six
from six.moves import range

# Define the logger
logger = logging.getLogger(__name__)

# Soma import
from soma.qt_gui import qt_backend
from soma.qt_gui.qt_backend import QtGui, QtCore
from soma.utils.functiontools import SomaPartial, partial
from soma.controller import trait_ids
from soma.controller import Controller
from soma.qt_gui.controller_widget import ControllerWidget, \
    ScrollControllerWidget, get_ref, weak_proxy

from .List import ListControlWidget, ListController
import traits.api as traits
import weakref
import sip

# Qt import
try:
    _fromUtf8 = QtCore.QString.fromUtf8
except AttributeError:
    _fromUtf8 = lambda s: s

QtCore.QResource.registerResource(os.path.join(os.path.dirname(
    os.path.dirname(__file__)), 'resources', 'widgets_icons.rcc'))


class OffscreenListControlWidget(object):

    """ Control to enter a list of items.
    """

    #
    # Public members
    #

    max_items = 5

    @staticmethod
    def is_valid(control_instance, *args, **kwargs):
        """ Method to check if the new control values are correct.

        If the new list controls values are not correct, the backroung
        color of each control in the list will be red.

        Parameters
        ----------
        control_instance: QFrame (mandatory)
            the control widget we want to validate

        Returns
        -------
        valid: bool
            True if the control values are valid,
            False otherwise
        """
        # Initilaized the output
        valid = True

        return valid

    @classmethod
    def check(cls, control_instance):
        """ Check if a controller widget list control is filled correctly.

        Parameters
        ----------
        cls: OffscreenListControlWidget (mandatory)
            an OffscreenListControlWidget control
        control_instance: QFrame (mandatory)
            the control widget we want to validate
        """
        pass

    @staticmethod
    def add_callback(callback, control_instance):
        """ Method to add a callback to the control instance when the list
        trait is modified

        Parameters
        ----------
        callback: @function (mandatory)
            the function that will be called when a 'textChanged' signal is
            emited.
        control_instance: QFrame (mandatory)
            the control widget we want to validate
        """
        pass

    @staticmethod
    def create_widget(parent, control_name, control_value, trait,
                      label_class=None, user_data=None):
        """ Method to create the list widget.

        Parameters
        ----------
        parent: QWidget (mandatory)
            the parent widget
        control_name: str (mandatory)
            the name of the control we want to create
        control_value: list of items (mandatory)
            the default control value
        trait: Tait (mandatory)
            the trait associated to the control
        label_class: Qt widget class (optional, default: None)
            the label widget will be an instance of this class. Its constructor
            will be called using 2 arguments: the label string and the parent
            widget.

        Returns
        -------
        out: 2-uplet
            a two element tuple of the form (control widget: ,
            associated labels: (a label QLabel, the tools QWidget))
        """
        # Get the inner trait: expect only one inner trait
        # note: trait.inner_traits might be a method (ListInt) or a tuple
        # (List), whereas trait.handler.inner_trait is always a method
        if len(trait.handler.inner_traits()) != 1:
            raise Exception(
                "Expect only one inner trait in List control. Trait '{0}' "
                "inner trait is '{1}'.".format(
                    control_name, trait.handler.inner_traits()))
        inner_trait = trait.handler.inner_traits()[0]

        # Create the widget
        frame = QtGui.QFrame()
        frame.setFrameShape(QtGui.QFrame.StyledPanel)
        frame.user_data = user_data
        layout = QtGui.QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        frame.setLayout(layout)
        #item = QtGui.QLabel('&lt;list of %s&gt;'
                            #% str(inner_trait.trait_type.__class__.__name__))
        #item.setTextInteractionFlags(QtCore.Qt.TextSelectableByKeyboard |
                                     #QtCore.Qt.TextSelectableByMouse)
        #item.setFrameStyle(QtGui.QFrame.StyledPanel | QtGui.QFrame.Sunken)
        #layout.addWidget(item)

        # Create tools to interact with the list widget: expand or collapse -
        # add a list item - remove a list item
        tool_widget = QtGui.QWidget(parent)
        layout.addWidget(tool_widget)

        layout = QtGui.QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(2)
        tool_widget.setLayout(layout)
        # Store some parameters in the list widget
        frame.inner_trait = inner_trait
        frame.trait = trait
        frame.connected = False
        if control_value is traits.Undefined:
            control_value = []
        elif not isinstance(control_value, (list, tuple)):
            # in nipype MultiPath, single values are not in a list
            control_value = [control_value]
        frame.trait_name = control_name

        # Create the label associated with the list widget
        control_label = trait.label
        if control_label is None:
            control_label = control_name
        if label_class is None:
            label_class = QtGui.QLabel
        if control_label is not None:
            label = label_class(control_label, parent)
        else:
            label = None
        frame.label_class = label_class

        # view the model
        items = OffscreenListControlWidget.partial_view_widget(
            weak_proxy(parent), weak_proxy(frame), control_value)
        layout.addWidget(items)
        #layout.addWidget(QtGui.QLabel('...'))
        frame.control_widget = items
        frame.controller = items.control_widget.controller
        frame.controller_widget = items.control_widget.controller_widget

        # Create the tool buttons
        edit_button = QtGui.QToolButton()
        layout.addWidget(edit_button)
        # Set the tool icons
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(_fromUtf8(":/soma_widgets_icons/add")),
            QtGui.QIcon.Normal, QtGui.QIcon.Off)
        edit_button.setIcon(icon)
        edit_button.setFixedSize(30, 22)

        #layout.addStretch(1)

        # Set some callback on the list control tools
        # Resize callback
        edit_hook = partial(
            OffscreenListControlWidget.edit_elements, weak_proxy(parent),
            weak_proxy(frame), weak_proxy(edit_button))
        edit_button.clicked.connect(edit_hook)

        return (frame, label)

    @staticmethod
    def partial_view_widget(controller_widget, parent_frame, control_value):
        widget = OneLineQTableWidget()
        widget.horizontalHeader().hide()
        widget.verticalHeader().hide()
        widget.max_items = OffscreenListControlWidget.max_items

        if control_value is traits.Undefined:
            control_value = []
        elif not isinstance(control_value, (list, tuple)):
            # in nipype MultiPath, single values are not in a list
            control_value = [control_value]

        control_widget, control_label = ListControlWidget.create_widget(
            controller_widget, parent_frame.trait_name, control_value,
            parent_frame.trait, parent_frame.label_class,
            max_items=widget.max_items, user_data=parent_frame.user_data)

        control_label[0].deleteLater()
        control_label[1].deleteLater()
        del control_label

        n = len(control_value)
        max_items = widget.max_items
        if max_items > 0:
            m = min(max_items, n)
            if n > max_items:
                widget.setColumnCount(m + 1)
            else:
                widget.setColumnCount(n)
        else:
            m = n
            widget.setColumnCount(n)
        clayout = control_widget.layout()
        widget.setRowCount(1)
        for i in range(m):
            litem = clayout.itemAtPosition(i + 1, 1)
            if litem is not None:
                item = litem.widget()
                widget.setCellWidget(0, i, item)
            if qt_backend.get_qt_backend() == 'PyQt5':
                widget.horizontalHeader().setSectionResizeMode(
                    i, QtGui.QHeaderView.Stretch)
            else:
                widget.horizontalHeader().setResizeMode(
                    i, QtGui.QHeaderView.Stretch)
        if n > m:
            widget.setCellWidget(0, max_items, QtGui.QLabel('...'))
            if qt_backend.get_qt_backend() == 'PyQt5':
                widget.horizontalHeader().setSectionResizeMode(
                    max_items, QtGui.QHeaderView.Fixed)
            else:
                widget.horizontalHeader().setResizeMode(
                    max_items, QtGui.QHeaderView.Fixed)

        widget.resizeRowToContents(0)
        #scroll_height = widget.findChildren(QtGui.QScrollBar)[-1].height()
        height = widget.rowHeight(0) \
            + sum(widget.getContentsMargins()[1:4:2]) # + scroll_height
        widget.setSizePolicy(QtGui.QSizePolicy.Expanding,
                             QtGui.QSizePolicy.Minimum)
        #widget.viewport().setSizePolicy(QtGui.QSizePolicy.MinimumExpanding,
                                        #QtGui.QSizePolicy.Fixed)
        #widget.viewport().setFixedHeight(widget.rowHeight(0))
        widget.setFixedHeight(height)

        widget.control_widget = control_widget
        control_widget.hide()

        return widget

    @staticmethod
    def update_controller(controller_widget, control_name, control_instance,
                          *args, **kwarg):
        """ Update one element of the controller.

        At the end the controller trait value with the name 'control_name'
        will match the controller widget user parameters defined in
        'control_instance'.

        Parameters
        ----------
        controller_widget: ControllerWidget (mandatory)
            a controller widget that contains the controller we want to update
        control_name: str(mandatory)
            the name of the controller widget control we want to synchronize
            with the controller
        control_instance: QFrame (mandatory)
            the instance of the controller widget control we want to
            synchronize with the controller
        """
        ListControlWidget.update_controller(
            controller_widget, control_name,
            control_instance.control_widget.control_widget, *args, **kwarg)

    @classmethod
    def update_controller_widget(cls, controller_widget, control_name,
                                 control_instance):
        """ Update one element of the list controller widget.

        At the end the list controller widget user editable parameter with the
        name 'control_name' will match the controller trait value with the same
        name.

        Parameters
        ----------
        controller_widget: ControllerWidget (mandatory)
            a controller widget that contains the controller we want to update
        control_name: str(mandatory)
            the name of the controller widget control we want to synchronize
            with the controller
        control_instance: QFrame (mandatory)
            the instance of the controller widget control we want to
            synchronize with the controller
        """
        try:
            if control_name not in controller_widget.controller.user_traits():
                return
        except ReferenceError:
            return

        # Get the list widget current connection status
        try:
            was_connected = control_instance.connected
            test = control_instance.control_widget.layout
        except (ReferenceError, RuntimeError):
            # widget deleted in the meantime
            return

        # Disconnect the list controller and the inner list controller
        cls.disconnect(controller_widget, control_name, control_instance)
        control_instance.controller_widget.disconnect()

        widget = control_instance.control_widget
        if sip.isdeleted(widget):
            OffscreenListControlWidget.disconnect(
                controller_widget, control_name, control_instance)
            return

        clayout = widget.control_widget.layout()
        owned_widgets = set()
        parent = widget
        for i in range(widget.columnCount()):
            w = widget.cellWidget(0, i)
            if w is not None:
                parent = w.parentWidget()
                owned_widgets.add(w)
        ListControlWidget.update_controller_widget(
            controller_widget, control_name, widget.control_widget)
        keys = list(control_instance.controller.user_traits().keys())
        #n = len(control_instance.controller.user_traits())
        parent_value = getattr(controller_widget.controller, control_name)
        if parent_value is traits.Undefined:
            parent_value = []
        elif not isinstance(parent_value, (list, tuple)):
            # in nipype MultiPath, single values are not in a list
            parent_value = [parent_value]
        n = len(parent_value)
        max_items = widget.max_items
        if max_items > 0:
            m = min(max_items, n)
            if n > max_items:
                widget.setColumnCount(m + 1)
            else:
                widget.setColumnCount(n)
        else:
            m = n
            widget.setColumnCount(n)
        for i in range(m):
            try:
                items = widget.control_widget.controller_widget._controls[
                    str(i)]
                item = items[None][2]
                if item not in owned_widgets:
                    widget.setCellWidget(0, i, item)
                if qt_backend.get_qt_backend() == 'PyQt5':
                    widget.horizontalHeader().setSectionResizeMode(
                        i, QtGui.QHeaderView.Stretch)
                else:
                    widget.horizontalHeader().setResizeMode(
                        i, QtGui.QHeaderView.Stretch)
            except KeyError:
                print('KeyError in OffscreenListControlWidget', i)
                print('controls:', widget.control_widget.controller_widget._controls)
        if n > m:
            label = QtGui.QLabel('...')
            width = label.sizeHint().width()
            widget.setCellWidget(0, max_items, QtGui.QLabel('...'))
            if qt_backend.get_qt_backend() == 'PyQt5':
                widget.horizontalHeader().setSectionResizeMode(
                    max_items, QtGui.QHeaderView.Fixed)
            else:
                widget.horizontalHeader().setResizeMode(
                    max_items, QtGui.QHeaderView.Fixed)
            widget.horizontalHeader().resizeSection(max_items, width)

        widget.resizeRowToContents(0)
        height = widget.rowHeight(0) \
            + sum(widget.getContentsMargins()[1:4:2])
        widget.setFixedHeight(height)

        # Restore the previous list controller connection status
        if was_connected:
            cls.connect(controller_widget, control_name, control_instance)
            control_instance.controller_widget.connect()

    @classmethod
    def connect(cls, controller_widget, control_name, control_instance):
        """ Connect a 'List' controller trait and an
        'OffscreenListControlWidget' controller widget control.

        Parameters
        ----------
        cls: StrControlWidget (mandatory)
            a StrControlWidget control
        controller_widget: ControllerWidget (mandatory)
            a controller widget that contains the controller we want to update
        control_name: str (mandatory)
            the name of the controller widget control we want to synchronize
            with the controller
        control_instance: QFrame (mandatory)
            the instance of the controller widget control we want to
            synchronize with the controller
        """
        # Check if the control is connected
        if not control_instance.connected:

            # Update the list item when one of his associated controller trait
            # changed.
            # Hook: function that will be called to update the controller
            # associated with a list widget when a list widget inner controller
            # trait is modified.
            list_controller_hook = SomaPartial(
                cls.update_controller, weak_proxy(controller_widget),
                control_name, weak_proxy(control_instance))

            # Go through all list widget inner controller user traits
            for trait_name in control_instance.controller.user_traits():

                # And add the callback on each user trait
                control_instance.controller.on_trait_change(
                    list_controller_hook, trait_name, dispatch='ui')
                logger.debug("Item '{0}' of a 'ListControlWidget', add "
                             "a callback on inner controller trait "
                             "'{0}'.".format(control_name, trait_name))

            # Update the list controller widget.
            # Hook: function that will be called to update the specific widget
            # when a trait event is detected on the list controller.
            controller_hook = SomaPartial(
                cls.update_controller_widget, weak_proxy(controller_widget),
                control_name, weak_proxy(control_instance))

            # When the 'control_name' controller trait value is modified,
            # update the corresponding control
            controller_widget.controller.on_trait_change(
                controller_hook, control_name, dispatch='ui')

            # Update the list connection status
            control_instance._controller_connections = (
                list_controller_hook, controller_hook)
            logger.debug("Add 'List' connection: {0}.".format(
                control_instance._controller_connections))

            # Update the list control connection status
            control_instance.connected = True

    @staticmethod
    def disconnect(controller_widget, control_name, control_instance):
        """ Disconnect a 'List' controller trait and an
        'OffscreenListControlWidget' controller widget control.

        Parameters
        ----------
        cls: StrControlWidget (mandatory)
            a StrControlWidget control
        controller_widget: ControllerWidget (mandatory)
            a controller widget that contains the controller we want to update
        control_name: str (mandatory)
            the name of the controller widget control we want to synchronize
            with the controller
        control_instance: QFrame (mandatory)
            the instance of the controller widget control we want to
            synchronize with the controller
        """
        # Check if the control is connected
        if control_instance.connected:

            # Get the stored widget and controller hooks
            (list_controller_hook,
             controller_hook) = control_instance._controller_connections

            # Remove the controller hook from the 'control_name' trait
            controller_widget.controller.on_trait_change(
                controller_hook, control_name, remove=True)

            # Remove the list controller hook associated with the controller
            # traits
            for trait_name in control_instance.controller.user_traits():
                control_instance.controller.on_trait_change(
                    list_controller_hook, trait_name, remove=True)

            # Delete the trait - control connection we just remove
            del control_instance._controller_connections

            # Disconnect also all list items
            inner_controls = control_instance.controller_widget._controls
            for (inner_control_name,
                 inner_control_groups) in six.iteritems(inner_controls):
                for group, inner_control \
                        in six.iteritems(inner_control_groups):

                    # Unpack the control item
                    inner_control_instance = inner_control[2]
                    inner_control_class = inner_control[1]

                    # Call the inner control disconnect method
                    inner_control_class.disconnect(
                        control_instance.controller_widget, inner_control_name,
                        inner_control_instance)

            # Update the list control connection status
            control_instance.connected = False

    #
    # Callbacks
    #


    @staticmethod
    def edit_elements(controller_widget, control_instance, edit_button):
        """ Callback to view/edit a 'ListControlWidget'.

        Parameters
        ----------
        control_instance: QFrame (mandatory)
            the list widget item
        edit_button: QToolButton
            the signal sender
        """
        controller_widget = get_ref(controller_widget)
        widget = QtGui.QDialog(controller_widget)
        widget.setModal(True)
        layout = QtGui.QVBoxLayout()
        widget.setLayout(layout)
        #hlayout = QtGui.QHBoxLayout()
        #layout.addLayout(hlayout)

        temp_controller = Controller()
        trait = control_instance.trait

        temp_controller.add_trait(control_instance.trait_name, trait)
        if temp_controller.trait(control_instance.trait_name).groups \
                is not None:
            temp_controller.trait(control_instance.trait_name).groups = None

        value = getattr(controller_widget.controller,
                        control_instance.trait_name)

        try:
            setattr(temp_controller, control_instance.trait_name, value)
        except Exception:
            # invalid value - don't prevent using the GUI
            pass
        control_types = dict(controller_widget._defined_controls)
        control_types['List'] = ListControlWidget
        control_types['List_File'] = ListControlWidget
        temp_controller_widget = ScrollControllerWidget(
            temp_controller, live=True, override_control_types=control_types,
            user_data=control_instance.user_data)

        layout.addWidget(temp_controller_widget)

        hlayout2 = QtGui.QHBoxLayout()
        layout.addLayout(hlayout2)
        hlayout2.addStretch(1)
        ok = QtGui.QPushButton('OK')
        cancel = QtGui.QPushButton('Cancel')
        hlayout2.addWidget(ok)
        hlayout2.addWidget(cancel)

        ok.pressed.connect(widget.accept)
        cancel.pressed.connect(widget.reject)

        if widget.exec_():

            ctrl = temp_controller_widget.controller_widget._controls.get(
                control_instance.trait_name)[None]
            ListControlWidget.validate_all_values(
                temp_controller_widget.controller_widget, ctrl[2])
            new_trait_value = getattr(temp_controller,
                                      control_instance.trait_name)

            setattr(controller_widget.controller,
                    control_instance.trait_name,
                    new_trait_value)

        del temp_controller_widget

    #@staticmethod
    #def item_changed(control_widget, item):
        #print('item_changed:', controller_widget, item)


class OneLineQTableWidget(QtGui.QTableWidget):
    '''QTableWidget witon one line, with fixed row height. Resizes to add the
    horizontal scrollbar at the bottom when it becomes visible without changing
    the visible row height.
    '''
    def resizeEvent(self, event):
        scrollbar = self.horizontalScrollBar().isVisible()
        mheight = self.rowHeight(0) \
            + sum(self.getContentsMargins()[1:4:2])
        if self.height() != mheight:
            self.setFixedHeight(mheight)
        super(OneLineQTableWidget, self).resizeEvent(event)



# -*- coding: utf-8 -*-
'''
Functions to manage private/public keys encryption.
This module needs Crypto module.
'''

from __future__ import absolute_import
from Crypto.PublicKey import RSA
from Crypto.Cipher import PKCS1_OAEP
from base64 import b64decode, b64encode
import Crypto
import six
from six.moves import range

if [int(x) for x in Crypto.__version__.split('.')] < [2, 1]:
    # pyrypro <= 2.0.x is not suitable since it is missing needed functions
    # such as exportKey() and importKey()
    raise ImportError('Crypto module (pycrypto) is too old for soma.crypt '
                      '- please install pycryptodome')


def generate_RSA(bits=2048):
    '''
    Generate an RSA keypair with an exponent of 65537 in PEM format
    param: bits The key length in bits
    Return private key and public key
    '''
    if [int(x) for x in Crypto.__version__.split('.')] < [2, 1]:
        import random
        def gen_func(n):
            return ''.join([chr(random.randrange(0, 256, 1))
                            for i in range(n)])
        new_key = RSA.generate(bits, gen_func)
    else:
        new_key = RSA.generate(bits)
    public_key = new_key.publickey().exportKey("PEM")
    private_key = new_key.exportKey("PEM")
    return private_key, public_key


def encrypt_RSA(public_key_loc, message):
    '''
    param: public_key_loc Path to public key
    param: message String to be encrypted
    return base64 encoded encrypted string
    '''
    key = open(public_key_loc, "rb").read()
    rsakey = RSA.importKey(key)
    if not isinstance(message, six.binary_type):
        message = message.encode()
    encryptor = PKCS1_OAEP.new(rsakey)
    encrypted = encryptor.encrypt(message)
    #encrypted = rsakey.encrypt(message, None)[0]
    return b64encode(encrypted)


def decrypt_RSA(private_key_loc, package):
    '''
    param: public_key_loc Path to your private key
    param: package String to be decrypted
    return decrypted string
    '''
    key = open(private_key_loc, "rb").read()
    rsakey = RSA.importKey(key)
    decryptor = PKCS1_OAEP.new(rsakey)
    decrypted = decryptor.decrypt(b64decode(package))
    #decrypted = rsakey.decrypt(b64decode(package))
    return decrypted



# -*- coding: utf-8 -*-

#  This software and supporting documentation are distributed by
#      Institut Federatif de Recherche 49
#      CEA/NeuroSpin, Batiment 145,
#      91191 Gif-sur-Yvette cedex
#      France
#
# This software is governed by the CeCILL-B license under
# French law and abiding by the rules of distribution of free software.
# You can  use, modify and/or redistribute the software under the
# terms of the CeCILL-B license as circulated by CEA, CNRS
# and INRIA at the following URL "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL-B license and that you accept its terms.

'''
This module provides a notification system that can be used to register
callbacks (*i.e* Python callables) that will all be called by a single
:meth:`Notifier.notify` call.

* author: Yann Cointepas, Dominique Geffroy
* organization: NeuroSpin
* license: `CeCILL B <http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html>`_
'''
from __future__ import absolute_import
__docformat__ = "restructuredtext en"

import six
from six.moves import range
import sys

from soma.translation import translate as _
from soma.functiontools import checkParameterCount, numberOfParameterRange
from soma.undefined import Undefined
from soma.sorted_dictionary import SortedDictionary

#-------------------------------------------------------------------------


class Notifier(object):

    '''
    Register a series of functions (or Notifier instances) which are all called
    whith the :meth:`notify` method. The calling order is the registering
    order. If a Notifier is registered, its :meth:`notify` method is called
    whenever *self.notify()* is called.
    '''

    def __init__(self, parameterCount=None):
        '''
        Parameters
        ----------
        parameterCount: int
            if not None, each registered function must be callable with that
            number of arguments (checking is done on registration).
        '''
        self._listeners = []
        self._parameterCount = parameterCount
        self._delayedNotification = None

    def add(self, listener):
        '''
        Register a callable or a Notifier that will be called whenever
        :meth:`notify` is called. If the notifier has a *parameterCount*,
        :func:`checkParameterCount <soma.utils.functiontools.checkParameterCount>` is used to verify that
        *listener* can be called with *parameterCount* parameters.

        Parameters
        ----------
        listener: Python callable (function, method, *etc*.) or :class:`Notifier` instance
            item to add to the notification list.
        '''

        if listener not in self._listeners:
            if self._parameterCount is not None:
                if isinstance(listener, Notifier):
                    if listener._parameterCount is not None and \
                       listener._parameterCount != self._parameterCount:
                        raise RuntimeError(_('Impossible to register a notifier with'
                                             '%(other)d parameter(s) to a notifier '
                                             'with %(self)d parameter(s)')
                                           % {'self': self._parameterCount,
                                              'other': listener._parameterCount})
                else:
                    checkParameterCount(listener, self._parameterCount)
            self._listeners.append(listener)

    def remove(self, listener):
        '''
        Remove an item from the notification list. Do nothing if *listener* is
        not in the list.

        Parameters
        ----------
        listener: Python callable or :class:`Notifier` instance
            item previously registered with :meth:`add`.
        Returns
        -------
        bool:
            *True* if a listener has been removed, *False* otherwise.
        '''

        try:
            self._listeners.remove(listener)
            result = True
        except ValueError:
            result = False
        return result

    def notify(self, *args):
        '''
        Calls all the registered items in the notification list. All the
        parameters given to :meth:`notify` are passed to the items in the list.
        For items in the list that are :class:`Notifier` instance, their
        :meth:`notify` method is called

        .. seealso:: :meth:`delayNotification`, :meth:`restartNotification`
        '''
        if self._delayedNotification is None:
            # if self._listeners: print '!notify!', self, ':', args, '(' + str(len( self._listeners )), 'listeners)'
            # Iterate on a copy of self._listeners because this list can be modified
            # by a listener during notification loop.
            for listener in tuple(self._listeners):
                # print '!notify!  -->', listener
                try:
                    if isinstance(listener, Notifier):
                        listener.notify(*args)
                    else:
                        listener(*args)
                except ReferenceError:
                    # listener is deleted in a weak ref/proxy
                    self._listeners.remove(listener)

        else:
            if self._delayedNotificationIgnoreDoubles:
                if args not in self._delayedNotification:
                    self._delayedNotification.append(args)
            else:
                self._delayedNotification.append(args)

    def delayNotification(self, ignoreDoubles=False):
        '''
        Stop notification until :meth:`restartNotification` is called. After a
        call to :meth`delayNotification`, all calls to :meth:`notify` will only
        store the notification parameters until :meth:`restartNotification` is
        called.

        Parameters
        ----------
        ignoreDoubles: bool
            If *True* (the default), all calls to :meth:`notify` with
            the same parameters as a previous call will be ignored (*i.e.*
            notification will be done only once for two identical calls).
        '''
        self._delayedNotification = []
        self._delayedNotificationIgnoreDoubles = ignoreDoubles

    def restartNotification(self):
        '''
        Restart notifications that have been delayed by
        :meth:`delayNotification`. All the calls to :meth:`notify` that have
        been done between the call to :meth:`delayNotification` and the call to
        :meth:`restartNotification`, are applied immediately.
        '''
        delayedNotification = self._delayedNotification
        if delayedNotification is not None:
            self._delayedNotification = None
            for args in delayedNotification:
                self.notify(*args)


#-------------------------------------------------------------------------
class ReorderedCall(object):

    '''
    **todo:** documentation
    '''

    def __init__(self, function, parametersOrder):
        self._function = function
        self._order = parametersOrder

    def __call__(self, *args):
        return self._function(*[args[i] for i in self._order])


#-------------------------------------------------------------------------
class VariableParametersNotifier(Notifier):

    '''
    This class is a notifier that can register functions with various arguments
    count.
    '''

    def __init__(self, mainParameters, *otherParameters):
        '''
        .. seealso:: :class:`Notifier`

        **todo** documentation
        '''
        Notifier.__init__(self, len(mainParameters))
        self.__parameters = {
            len(mainParameters): list(range(len(mainParameters)))
        }
        self.__min = len(mainParameters)
        self.__max = len(mainParameters)
        # Converting to list because it has the index method (which tuple has
        # not)
        mainParameters = list(mainParameters)
        main = set(mainParameters)
        for p in otherParameters:
            other = set(p)
            if other - main:
                raise RuntimeError(_('Invalid parameters definition (%(other)s is'
                                     ' not a subset of %(main)s)') %
                                   {'other': str(other), 'main': str(main)})
            if len(p) in self.__parameters:
                raise RuntimeError(_('Invalid parameters definition (several sets'
                                     ' of %d parameters defined)') % (len(p),))
            self.__parameters[len(p)] = [
                mainParameters.index(i) for i in p]
            self.__min = min(self.__min, len(mainParameters))

    def add(self, listener):
        '''
        .. seealso:: :meth:`Notifier.add`

        **todo:** documentation
        '''
        if isinstance(listener, Notifier):
            realListener = listener
        else:
            min, max = numberOfParameterRange(listener)
            if max is None:
                paramCount = self.__max
            else:
                paramCount = max
            paramOrder = self.__parameters.get(paramCount)
            if paramOrder is None:
                raise RuntimeError(_('%(f)s has an invalid parameter count '
                                     '(%(c)d)') %
                                   {'f': str(listener), 'c': paramCount})
            realListener = ReorderedCall(listener, paramOrder)
        Notifier.add(self, realListener)

    def remove(self, listener):
        '''
        .. seealso:: :meth:`Notifier.remove`

        **todo:** documentation
        '''

        for i in range(len(self._listeners)):
            c = self._listeners[i]
            if ( isinstance( c, ReorderedCall ) and c._function == listener ) or \
                    c == listener:
                del self._listeners[i]
                return True
        return False


#-------------------------------------------------------------------------
class ObservableAttributes(object):

    '''
    ObservableAttributes allows to track modification of attributes at
    runtime. By registering callbacks, it is possible to be warn of the
    modification (setting and deletion) of any attribute of the instance.
    '''

    def __init__(self, *args, **kwargs):

        #: VariableParametersNotifier instance notified whenever any attribute
        #: is modified. Use self.onAttributeChange to register a function on
        #: this notifier.
#    self.__dict__[ '_onAnyAttributeChange' ] = \
#      self._createAttributeNotifier()
        super(ObservableAttributes, self).__setattr__('_onAnyAttributeChange',
                                                      self._createAttributeNotifier())

        #: Dictionary whose keys are attribute names and values are
        #: L{VariableParametersNotifier} instances. Whenever an attribute is
        #: modified, the corresponding L{VariableParametersNotifier} is notified.
        #: Use L{self.onAttributeChange} to register a function on these notifiers.
#    self.__dict__[ '_onAttributeChange' ] = {}
        super(ObservableAttributes, self).__setattr__(
            '_onAttributeChange', {})
        super(ObservableAttributes, self).__init__(*args, **kwargs)

    @staticmethod
    def _createAttributeNotifier():
        '''
        Static method creating a :class:`VariableParametersNotifier` instance
        for attribute modification notification.

        see :meth:`notifyAttributeChange`.
        '''
        return VariableParametersNotifier(
            ('object', 'attributeName', 'newValue', 'oldValue'),
            (),
            ('newValue', ),
            ('newValue', 'oldValue'),
            ('attributeName', 'newValue', 'oldValue'),
        )

    def notifyAttributeChange(self, name, value, oldValue=Undefined):
        '''
        First, calls functions registered for modification of the attribute
        named *name*, then call functions registered for modification of any
        attribute.

        .. seealso:: :meth:`onAttributeChange`
        '''
        # Notify the change of this attribute
        if hasattr(self, '_onAttributeChange'):
            notifier = self._onAttributeChange.get(name)
            if notifier is not None:
                notifier.notify(self, name, value, oldValue)

        if hasattr(self, '_onAnyAttributeChange'):
            # Notify the change of one attribute
            self._onAnyAttributeChange.notify(self, name, value, oldValue)

    def __setattr__(self, name, value):
        '''
        Changes the value of attribute *name* then calls
        :meth:`notifyAttributeChange`.
        '''
        oldValue = getattr(self, name, Undefined)
        super(ObservableAttributes, self).__setattr__(name, value)
        if value != oldValue:
            self.notifyAttributeChange(name, value, oldValue)

    def __delattr__(self, name):
        '''
        Deletes the attribute *name* then calls :meth:`notifyAttributeChange`
        with ``newValue = :class:`Undefined` ``.
        '''
        oldValue = getattr(self, name, Undefined)
        super(ObservableAttributes, self).__delattr__(name)
        self.notifyAttributeChange(name, Undefined, oldValue)
        # Delete notifier for the deleted attribute
        self._onAttributeChange.pop(name, None)

    def onAttributeChange(self, first, second=None):
        '''
        Registers a function to be called when an attribute is modified or
        deleted. To call the function for any attribute modification, use the
        following syntax::

            instance.onAttributeChange(function)

        To register a function for a named attribute, use the following
        syntax::

            instance.onAttributeChange(attributeName, function)

        The registered function can have 0 to 4 parameters. Depending on its
        number of parameters, it will be called with the following values:

        - 4 parameters: (object, attributeName, newValue, oldValue)
        - 3 parameters: (attributeName, newValue, oldValue)
        - 2 parameters: (newValue, oldValue)
        - 1 parameter: (newValue)

        Where:

        - **object** is the object whose attribute has been modified or
          deleted.
        - **attributeName** is the name of the modified or deleted attribute.
        - **newValue** is the value of the attribute after modification or
          ``Undefined`` if the attribute has been deleted.
        - **oldValue** is the value of the attribute before modification. If
          the attribute was not defined, ``oldValue = Undefined``.

        If the function accepts a variable number of parameters, it will be
        called with the maximum number of arguments possible.
        '''
        if second is None:
            if hasattr(self, '_onAnyAttributeChange'):
                self._onAnyAttributeChange.add(first)
        else:
            if hasattr(self, '_onAttributeChange'):
                notifier = self._onAttributeChange.get(first)
                if notifier is None:
                    notifier = self._createAttributeNotifier()
                    self._onAttributeChange[first] = notifier
                notifier.add(second)

    def removeOnAttributeChange(self, first, second=None):
        '''
        Remove a function previously registered with :meth:`onAttributeChange`.
        To remove a function, the arguments of :meth:`removeOnAttributeChange`
        must be the same as those passed to :meth:`onAttributeChange` to
        register the function.
        '''
        if second is None:
            result = self._onAnyAttributeChange.remove(first)
        else:
            result = self._onAttributeChange[first].remove(second)
        return result

    def delayAttributeNotification(self, ignoreDoubles=False):
        '''
        Stop attribute modification notification until
        :meth:`restartAttributeNotification` is called. After a call to
        :meth:`delayAttributeNotification`, all modification notification will
        only be stored until :meth:`restartAttributeNotification` is called.
        This call is recursive on all attributes values that are instance of
        :class:`ObservableAttributes`.

        Parameters
        ----------
        ignoreDoubles: bool
            If True (False is the default), all notification with the same
            parameters as a previous notification will be ignored (*i.e.* notification will be done only once for two identical events).
        '''
        self._delayAttributeNotification(
            ignoreDoubles=ignoreDoubles, checkedObjects=set())

    def _delayAttributeNotification(self, ignoreDoubles=False,
                                    checkedObjects=None):

        if not checkedObjects == None:
            checkedObjects.add(self)

        for name, notifier in six.iteritems(self._onAttributeChange):
            notifier.delayNotification(ignoreDoubles)
        self._onAnyAttributeChange.delayNotification(ignoreDoubles)

        # Recursively delay notification
        for name in dir(self):
            value = getattr(self, name, Undefined)

            if isinstance(value, ObservableAttributes):
                if not checkedObjects is None:
                    # This allow to not recursively call
                    # _delayAttributeNotification
                    if not value in checkedObjects:
                        value._delayAttributeNotification(
                            ignoreDoubles=ignoreDoubles, checkedObjects=checkedObjects)
                else:
                    value._delayAttributeNotification(
                        ignoreDoubles=ignoreDoubles)

        if not checkedObjects == None:
            checkedObjects.pop()

    def restartAttributeNotification(self):
        '''
        Restarts notifications that have been delayed by
        :meth:`delayAttributeNotification`. All the modifications that happened
        between the call to :meth:`delayAttributeNotification` and the call to
        :meth:`restartAttributeNotification`, are notified immediately.
        '''
        self._restartAttributeNotification(checkedObjects=set())

    def _restartAttributeNotification(self, checkedObjects=None):

        if not checkedObjects == None:
            checkedObjects.add(self)

        for name, notifier in six.iteritems(self._onAttributeChange):
            notifier.restartNotification()
        self._onAnyAttributeChange.restartNotification()
        # Recursively restart notification
        for name in dir(self):
            value = getattr(self, name, Undefined)
            if isinstance(value, ObservableAttributes):
                if not checkedObjects is None:
                    # This allow to not recursively call
                    # _delayAttributeNotification
                    if not value in checkedObjects:
                        value.restartAttributeNotification()
                else:
                    value._delayAttributeNotification()

        if not checkedObjects == None:
            checkedObjects.pop()

#----------------------------------------------------------------------------


class ObservableList(list):

    """
    A list that notifies its changes to registred listeners.
    Inherits from python list and contains an instance of :class:`Notifier`
    (:attr:`onChangeNotifier`).

    Example::

        l = ObservableList()
        l.addListener(update)
        l.append(e)

    * calls :meth:`onChangeNotifier.notify(INSERT_ACTION, [e], len(l)) <Notifier.notify>`
    * calls ``update(INSERT_ACTION, [e], len(l))``

    Attributes
    ----------
    INSERT_ACTION: int
        used to notify insertion of new elements  in the list
    REMOVE_ACTION: int
        used to notify elements deletion
    MODIFY_ACTION: int
        used to notify elements modification

    onChangeNotifier: Notifier
        the Notifier's notify method is called when the list
        has changed.
    """

    # actions to notify
    INSERT_ACTION = 0
    REMOVE_ACTION = 1
    MODIFY_ACTION = 2

    def __init__(self, content=None):
        """
        Parameters
        ----------
        content: list
            elements to initialize the list content
        """
        # call a super class method can be done two different ways:
        # - superClass.method(self, ...)
        # - super(superClass, self).method(...)
        # It's better to use the second way because if derived class inherits from several classes,
        # super will try to find the method in other super classes before
        super(ObservableList, self).__init__()
        # views can register update callbacks on this notifier
        # to be aware of any change in the model
        # On change, this object calls Notifier.notify(args)
        # which calls every registred function with args
        # args = action (insert, remove, modify), elems list, position
        self.onChangeNotifier = Notifier()
        if content:
            self.extend(content)

    def __getinitargs__(self):
        """Returns the args to pass to the __init__ method to construct this
        object.
        It is useful to save an :class:`ObservableList` object to :mod:`minf`
        format.

        Returns
        -------
        tuple:
            arg content to pass to the __init__ method for creating a copy of this object
        """
        content = []
        content.extend(self)
        return (content)

    def __reduce__(self):
        """This method is redefined for enable deepcopy of this object (and
        potentially pickle).
        It gives the arguments to pass to the init method of the object when
        creating a copy

        Returns
        -------
        tuple:
            class name, init args, state, iterator on elements to copy,
            dictionary iterator
        """
        # class name, init args, parameters for setstate, elements iterator,
        # dictionary iterator
        return (self.__class__, (), None, iter(self), None)

    def addListener(self, listener):
        """Registers the listener callback method in the notifier.
        The method must take 3 arguments: action, elems list, position

        *action* should be one of:

        - INSERT_ACTION: elems have been inserted at position in the list
        - REMOVE_ACTION: elems have been removed [at position] in the list
        - MODIFY_ACTION: at position, some elements have been replaced by
          elems

        The position given in the notify method will be between 0 and
        ``len(self)``

        Parameters
        ----------
        listener: function
            function to call to notify changes
        """
        self.onChangeNotifier.add(listener)

    def append(self, elem):
        """Adds the element at the end of the list.
        Notifies an insert action.
        """
        index = len(self)
        super(ObservableList, self).append(elem)
        self.onChangeNotifier.notify(self.INSERT_ACTION, [elem], index)

    def extend(self, l):
        """Adds the content of the list l at the end of current list.
        Notifies an insert action. """
        index = len(self)
        super(ObservableList, self).extend(l)
        self.onChangeNotifier.notify(self.INSERT_ACTION, l, index)

    def insert(self, pos, elem):
        """Inserts elem at position pos in the list.
        Notifies an insert action.
        """
        index = self.getPositiveIndex(pos)
        super(ObservableList, self).insert(pos, elem)
        self.onChangeNotifier.notify(self.INSERT_ACTION, [elem], index)

    def remove(self, elem):
        """Removes the first occurence of elem in the list.

        Notifies a remove action. """
        super(ObservableList, self).remove(elem)
        self.onChangeNotifier.notify(self.REMOVE_ACTION, [elem])

    def pop(self, pos=None):
        """Removes the element at position pos or the last element if pos is
        *None*.

        Notifies a remove action.

        Returns
        -------
        object:
            the removed element
        """
        if pos is not None:
            index = self.getPositiveIndex(pos)
            elem = super(ObservableList, self).pop(pos)
        else:
            index = len(self) - 1
            elem = super(ObservableList, self).pop()
        self.onChangeNotifier.notify(self.REMOVE_ACTION, [elem], index)
        return elem

    def sort(self, key=None, reverse=False):
        """Sorts the list using key function key.

        Notifies a modify action.

        Returns
        -------
        key: function
            key function: elem->key
        """
        super(ObservableList, self).sort(key=key, reverse=reverse)
        # all the elements of the list could be modified
        self.onChangeNotifier.notify(self.MODIFY_ACTION, self, 0)

    def reverse(self):
        """Inverses the order of the list.
        Notifies a modify action."""
        super(ObservableList, self).reverse()
        self.onChangeNotifier.notify(self.MODIFY_ACTION, self, 0)

    def __setitem__(self, key, value):
        """Sets value to element at position key in the list.

        Notifies a modify action::

            l[key] = value
        """
        index = self.getPositiveIndex(key)
        super(ObservableList, self).__setitem__(key, value)
        self.onChangeNotifier.notify(self.MODIFY_ACTION, [value], index)

    def __delitem__(self, key):
        """Removes the element at position key in the list.

        Notifies a remove action::

            del l[key]
        """
        index = self.getPositiveIndex(key)
        super(ObservableList, self).__delitem__(key)
        self.onChangeNotifier.notify(self.REMOVE_ACTION, [], index)

    def __setslice__(self, i, j, seq):
        """Sets values in seq to elements in the interval i,j.

        If i and j are negative numbers, there are converted before this call
        in ``index + len(self)``

        Notifies a modify action:

            l[i:j] = seq
        """
        indexI = self.getIndexInRange(i)
        indexJ = self.getIndexInRange(j)
        super(ObservableList, self).__setslice__(i, j, seq)
        # if the interval is empty, action is insertion at the first position
        if indexI >= indexJ:
            self.onChangeNotifier.notify(self.INSERT_ACTION, seq, indexI)
        else:
            lenSeq = len(seq)
            lenInter = indexJ - indexI
            # if values sequence has same or lower length as the interval in the list,
            # all values are written to the list from the first index
            # the rest of the interval (if interval is longer than sequence) is
            # left unchanged
            if lenInter >= lenSeq:
                self.onChangeNotifier.notify(self.MODIFY_ACTION, seq, indexI)
            else:
                # if the interval is shorter than the sequence of values,
                # values in the interval are used to modify the list,
                # the rest is inserted at indexJ position
                self.onChangeNotifier.notify(
                    self.MODIFY_ACTION, seq[0:lenInter], indexI)
                self.onChangeNotifier.notify(
                    self.INSERT_ACTION, seq[lenInter:lenSeq], indexJ)

    def __delslice__(self, i, j):
        """Removes elements in the interval i,j.

        If i and j are negative numbers, there are converted before this call
        in ``index + len(self)``.

        Notifies a remove action::

            del l[i:j]
        """
        indexI = self.getIndexInRange(i)
        indexJ = self.getIndexInRange(j)
        seq = self[indexI:indexJ]
        super(ObservableList, self).__delslice__(i, j)
        # if the interval is empty, the list is not modified
        if indexI < indexJ:
            self.onChangeNotifier.notify(self.REMOVE_ACTION, seq, indexI)

    def __iadd__(self, l):
        """``list += l`` <=> ``list.extend(l)``

        Notifies insert action."""
        index = len(self)
        newList = super(ObservableList, self).__iadd__(l)
        self.onChangeNotifier.notify(self.INSERT_ACTION, l, index)
        return newList

    def __imul__(self, n):
        """``list *= n``

        Notifies insert action."""
        index = len(self)
        newList = super(ObservableList, self).__imul__(n)
        self.onChangeNotifier.notify(self.INSERT_ACTION, self[index:], index)
        return newList

    def getPositiveIndex(self, i):
        """Returns an index between 0 and ``len(self)``

        - if *i* is negative, it is replaced by ``len(self) + i``
        - if *index* is again negative, it is replaced by 0
        - if *index* is beyond ``len(self)`` it is replaced by ``len(self)``
        """
        l = len(self)
        if i < 0:
            index = max(0, l + i)
        else:
            index = min(i, l)
        return index

    def getIndexInRange(self, i):
        """Returns an index in the range of the list.

        * if ``i < 0`` returns 0
        * if ``i > len(self)``, returns ``len(self)``
        """
        if i < 0:
            index = 0
        else:
            l = len(self)
            if i > l:
                index = l
            else:
                index = i
        return index

    def itemIndex(self, item):
        """
        Returns item's index in the list. It is different from the
        :meth:`list.index` method that returns the index of the first element
        that has the same value as *item*.
        """
        i = 0
        for it in self:
            if it is item:
                break
            i = i + 1
        if i == len(self):
            i = -1
        return i

#----------------------------------------------------------------------------


class ObservableSortedDictionary(SortedDictionary):

    """
    A sorted dictionary that notifies its changes.
    Inherits from python list and contains an instance of
    :class:`Notifier` (:attr:`onChangeNotifier`).

    Example::

        d=ObservableSortedDictionary()
        d.addListener(update)
        d.insert(index, key, e)

    * calls :meth:`onChangeNotifier.notify(INSERT_ACTION, [e], index) <Notifier.notify>`
    * calls ``update(INSERT_ACTION, [e], index)``

    Attributes
    ----------
    INSERT_ACTION: int
        used to notify insertion of new elements  in the dictionary
    REMOVE_ACTION: int
        used to notify elements deletion
    MODIFY_ACTION: int
        used to notify elements modification

    onChangeNotifier: Notifier
        the Notifier's notify method is called when the dictionaty has changed.
    """

    # actions to notify
    INSERT_ACTION = 0
    REMOVE_ACTION = 1
    MODIFY_ACTION = 2

    def __init__(self, *args):
        '''
        Initialize the dictionary with a list of ( key, value ) pairs.
        '''
        self.onChangeNotifier = Notifier()
        super(ObservableSortedDictionary, self).__init__(*args)

    def __getinitargs__(self):
        """Returns the args to pass to the __init__ method to construct this object.
        It is useful to save ObservableList object to minf format.
        Returns
        -------
        tuple:
            arg content to pass to the __init__ method for creating a copy of
            this object
        """
        content = list(self.items())
        return (content)

    def addListener(self, listener):
        """Registers the listener callback method in the notifier.
        The method must take 3 arguments: action, elems list, position

        *action* should be one of:

        - INSERT_ACTION: elems have been inserted at position in the dictionary
        - REMOVE_ACTION: elems have been removed [at position] in the dict
        - MODIFY_ACTION: at position, some elements have been replaced by elems

        The position given in the notify method will be between 0 and len(self)

        Parameters
        ----------
        listener: function
            function to call to notify changes
        """
        self.onChangeNotifier.add(listener)

    def __setitem__(self, key, value):
        insertion = key not in self
        super(ObservableSortedDictionary, self).__setitem__(key, value)
        if insertion:
            self.onChangeNotifier.notify(
                self.INSERT_ACTION, [value], len(self) - 1)
        else:
            self.onChangeNotifier.notify(
                self.MODIFY_ACTION, [value], self.sortedKeys.index(key))

    def __delitem__(self, key):
        index = self.sortedKeys.index(key)
        super(ObservableSortedDictionary, self).__delitem__(key)
        self.onChangeNotifier.notify(self.REMOVE_ACTION, [], index)

    def insert(self, index, key, value):
        '''
        inserts a (*key*, *value*) pair in the sorted dictionary before
        position *index*. If *key* is already in the dictionary, a
        :class:`KeyError <exceptions.KeyError>` is raised.

        Parameters
        ----------
        key:
            key to insert
        value:
            value associated to *key*

        Returns
        -------
        index: integer
            index of C{key} in the sorted keys
        '''
        super(ObservableSortedDictionary, self).insert(index, key, value)
        self.onChangeNotifier.notify(self.INSERT_ACTION, [value], index)

    def clear(self):
        '''
        Removes all items from dictionary
        '''
        super(ObservableSortedDictionary, self).clear()
        self.onChangeNotifier.notify(self.REMOVE_ACTION, list(self.values()), 0)

    def sort(self, key=None, reverse=False):
        """Sorts the dictionary using function *key* to compare keys.

        Notifies a modify action.

        Parameters
        ----------
        key: function
            key function key->key
        """
        super(ObservableSortedDictionary, self).sort(key=key, reverse=reverse)
        self.onChangeNotifier.notify(self.MODIFY_ACTION, list(self.values()), 0)


#----------------------------------------------------------------------------
class EditableTree(ObservableAttributes, ObservableSortedDictionary):

    """The base class to model a tree of items.
    This class can be derived to change implementation.

    An EditableTree contains items which can be

    - an item branch: it contains other items as children
    - an item leaf: doesn't have children

    The list of items is an :class:`ObservableSortedDictionary` which notifies
    its changes to registred listeners.

    If the tree is modifiable, new items can be added.

    Every item is an :class:`EditableTree.Item`.

    *EditableTree* is iterable over its items.

    *EditableTree* also inherits from :class:`ObservableAttributes`, so
    registred listeners can be notified of attributes value change.

    To call a method when item list changes::

      editableTree.addListener(callbackMethod)

    To call a method when an item list changes at any depth in the tree::

        editableTree.addListenerRec(callbackMethod)

    To call a method when an attribute of the tree changes::

        editableTree.onAttributeChange(attributeName, callbackMethod)

    To call a method when an attribute changes at any depth in the tree::

        editableTree.onAttributeChangeRec(attributeName, callbackMethod)

    Attributes
    ----------
    defaultName: string
        default name of the tree
    name: string
        the name of the tree
    id: string
        tree identifier (a hash by default)
    modifiable: bool
        if *True*, new items can be added, items can be deleted and modified
    unamed: bool
        indicates if *name* parameter was none, so the tree has the default
        name.
    visible: bool
        indicates if the tree is visible (if not it may be hidden in a
        graphical representation)
    """
    defaultName = "tree"

    def __init__(self, name=None, id=None, modifiable=True, content=[],
                 visible=True, enabled=True):
        """
        Parameters
        ----------
        name: string
            the name of the tree
        id: string
            tree identifier (a hash by default)
        modifiable: bool
            if *True*, new items can be added, items can be deleted and
            modified
        content: list
            children items (:class:`EditableTree.Item`)
        visible: bool
            indicates if the tree is visible (if not it may be hidden in a
            graphical representation)
        enabled: bool
        """
        dictContent = [(i.id, i) for i in content]
        super(EditableTree, self).__init__(*dictContent)
        if name is None:
            self.name = self.defaultName
            self.unamed = True
        else:
            self.name = name
            self.unamed = False
        if id is None:
            self.id = str(hash(self))
        else:
            self.id = id
        self.modifiable = modifiable
        self.visible = visible
        self.enabled = enabled

    def __getinitargs__(self):
        """Returns the args to pass to the __init__ method to construct this object.
        It is useful in order to save EditableTree object to minf format.
        Returns
        -------
        tuple:
            arg content to pass to the __init__ method for creating a copy of this object
        """
        # elements in the tuple must be serializable with minf, so the content
        # must be of type list
        content = list(self.values())
        return (self.name, self.id, self.modifiable, content, self.visible, self.enabled)

    def __str__(self):
        s = self.name + " ("
        for i in self:
            s += str(i) + " "
        s += ")"
        return s

    def __hash__(self):
        return ObservableAttributes.__hash__(self)

    def add(self, item):
        """
        Adds an item in the tree. If this item's id is already present in the
        tree as a key, add the item's content in the corresponding key.

        recursive method
        """
        key = item.id
        if key in self:
            if not item.isLeaf():  # if the item is a leaf and is already in the tree, nothing to do
                for v in item.values():  # item is also a dictionary and contains several elements, add each value in the tree item
                    self[key].add(v)
            # also set current name for the current object
            self[key].name = item.name
        else:  # new item
            self[key] = item

    def isDescendant(self, item):
        """Returns *True* if the current item is a child or a child's child...
        of the item in parameter

        Always *False* for the root of the tree"""
        return False

    def isLeaf(self):
        return False

    def removeEmptyBranches(self):
        """If a branch item doesn't contain any leaf or branch that contains
        leaf recursivly, the item is removed from the tree. """
        toRemove = []
        for child in self.values():
            if not child.isLeaf():
                child.removeEmptyBranches()
                if list(child.values()) == []:  # it isn't possible to remove an element during iteration on the list
                    toRemove.append(
                        child)  # so it's put on a remove list and will be removed from the tree outside the loop
        for item in toRemove:
            del self[item.id]

    def sort(self, key=None, reverse=False):
        """Recursive sort of the tree: items are sorted in all branches.
        """
        ObservableSortedDictionary.sort(self, self._keyItems)
        for item in self.values():
            if not item.isLeaf():
                item.sort(key=key, reverse=reverse)

    def _keyItems(self, id):
        """Key function
        """
        i1 = self[id]
        # names are translated in lowercase to make an alphabetical sort independant of the case
        # by default (uppercase letters are < to lowecase letters)
        n1 = i1.name.lower()
        # print "comp", i1.name,i1.isLeaf(), i2.name,i2.isLeaf(), res
        return (not i1.isLeaf(), n1)

    def compItems(self, id1, id2):
        """Comparison function

        Returns:

        - 1 if ``i1 > i2``
        - -1 if ``i1 < i2``
        - 0 if they are equal"""
        res = 0
        i1 = self[id1]
        i2 = self[id2]
        if (not i1.isLeaf() and i2.isLeaf()):
            # if one item is a branch and the other is a leaf, the branch must
            # be before the leaf
            res = -1
        elif (i1.isLeaf() and not i2.isLeaf()):
            res = 1
        else:
            # names are translated in lowercase to make an alphabetical sort independant of the case
            # by default (uppercase letters are < to lowecase letters)
            n1 = i1.name.lower()
            n2 = i2.name.lower()
            if n1 < n2:
                res = -1
            elif n1 > n2:
                res = 1
        # print "comp", i1.name,i1.isLeaf(), i2.name,i2.isLeaf(), res
        return res

    def addListenerRec(self, listener):
        """Add a listener to the tree recursivly at every level.
        The listener is added to the notifier of all branches of the tree."""
        self.addListener(listener)
        for item in self.values():
            if not item.isLeaf():
                item.addListenerRec(listener)

    def onAttributeChangeRec(self, attributeName, listener):
        """Add a listener of the changes of this attribute value in the tree
        recursivly at every level.

        The listener is added to the attribute change notifier of all branches
        of the tree.
        """
        self.onAttributeChange(attributeName, listener)
        for item in self.values():
            if not item.isLeaf():
                item.onAttributeChangeRec(attributeName, listener)

    #-------------------------------------------------------------------------
    class Item(ObservableAttributes):

        """Base element of an :class:`EditableTree`

        *Item* inherits from :class:`ObservableAttributes`, so it can notify
        registred listeners of attributes changes (for example :attr:`name`).

        Attributes
        ----------
        icon: string
            filename of the image that can be used to represent the item
        name: string
            Text representation of the item
        movable: bool
            True if the item can be moved in the tree
        modifiable: bool
            True if the item can changed (add children, change text, icon)
        delEnabled: bool
            True if deletion of the item is allowed
        tooltip: string
            description associated to the item
        copyEnabled: bool
            indicates if this item can be copied
        visible: bool
            indicates if current item is visible (if not it may be hidden in a
            graphical representation)
        enabled: bool
            indicates if the item is enabled, it could be visible but disabled

        """

        def __init__(self, name=None, id=None, icon=None, tooltip=None, copyEnabled=True, modifiable=True, delEnabled=True, visible=True, enabled=True, *args):
            super(EditableTree.Item, self).__init__(*args)
            self.icon = icon
            self.name = name
            if id is None:
                self.id = str(hash(self))
            else:
                self.id = id
            self.tooltip = tooltip
            self.copyEnabled = copyEnabled
            self.modifiable = modifiable
            self.delEnabled = delEnabled
            self.visible = visible
            self.enabled = enabled
            self.onChangeNotifier = Notifier()
            self.unamed = False

        def __getinitargs__(self):
            """Returns the args to pass to the __init__ method to construct this object.
            It is useful to save L{ObservableList} object to minf format.
            Returns
            -------
            tuple:
                arg content to pass to the __init__ method for creating a copy
                of this object
            """
            return (self.name, self.id, self.icon, self.tooltip, self.copyEnabled, self.modifiable, self.delEnabled, self.visible, self.enabled)

        def __reduce__(self):
            """This method is redefined for enable deepcopy of this object (and
            potentially pickle).

            It gives the arguments to pass to the init method of the object when creating a copy

            Returns
            -------
            tuple:
                class name, init args, state, iterator on elements to copy,
                dictionary iterator
            """
            return (self.__class__, self.__getinitargs__(), None, None, None)

        def __str__(self):
            return self.name

        def isLeaf(self):
            """Must be redefined in subclasses to say if the item is a leaf"""
            pass

        def setAllModificationsEnabled(self, bool):
            """Recursivly enables or disables item's modification."""
            # self.copyEnabled=bool # copy is always enabled. there's no use to
            # have 3 different booleans, an editable should be sufficient
            self.modifiable = bool
            self.delEnabled = bool
            if not self.isLeaf():
                for child in self.values():
                    child.setAllModificationsEnabled(bool)

        def isDescendant(self, item):
            """Returns *True* if the current item is a child or a child's
            child... of the item in parameter
            """
            found = False
            # print self.name, "is descendant", item.name, "? "
            if not item.isLeaf():
                for child in item.values():
                    # search self in item's children and recursively in
                    # children's children (deep first search)
                    if self == child:
                        found = True
                    elif not child.isLeaf():
                        found = self.isDescendant(child)
                    if found:
                        break
            return found

    #-------------------------------------------------------------------------
    class Branch(Item, ObservableSortedDictionary):

        """ A Branch is an :class:`Item <EditableTree.Item>` that can contain
        other items.

        It inherits from :class:`Item <EditableTree.Item>` and from
        :class:`ObservableSortedDictionary`, so it can have children items.
        """
        defaultName = "new"

        def __init__(self, name=None, id=None, icon=None, tooltip=None, copyEnabled=True, modifiable=True, delEnabled=True, content=[], visible=True, enabled=True):
            """All parameters must have default values to be able to create new
            elements automatically"""
            # super(EditableTree.Branch, self).__init__(content)  #, name, icon, tooltip, copyEnabled, modifiable, delEnabled)
            # EditableTree.Item.__init__(self, name, icon, tooltip,
            # copyEnabled, modifiable, delEnabled)
            dictContent = [(i.id, i) for i in content]
            super(EditableTree.Branch, self).__init__(name, id, icon, tooltip,
                                                      copyEnabled, modifiable, delEnabled, visible, enabled, *dictContent)
            if name is None:
                self.name = self.defaultName
                self.unamed = True
            else:
                self.name = name
                self.unamed = False

        def __getinitargs__(self):
            content = list(self.values())
            return (self.name, self.id, self.icon, self.tooltip, self.copyEnabled, self.modifiable, self.delEnabled, content, self.visible, self.enabled)

        def __reduce__(self):
            """This method is redefined for enable deepcopy of this object (and
            potentially pickle).
            It gives the arguments to pass to the init method of the object
            when creating a copy
            """
            return (self.__class__, self.__getinitargs__(), None, None, None)

        def __str__(self):
            s = self.name + " ("
            for i in self:
                s += str(i) + " "
            s += ")"
            return s

        def __hash__(self):
            return EditableTree.Item.__hash__(self)

        def isLeaf(self):
            return False

        def add(self, item):
            """
            Adds an item in the tree. If this item's id is already present in
            the tree as a key, add the item's content in the corresponding key.

            recursive method
            """
            key = item.id
            if key in self:
                if not self[key].isLeaf():  # if the item is a leaf and is already in the tree, nothing to do
                    for v in item.values():  # item is also a dictionary and contains several elements, add each value in the tree item
                        self[key].add(v)

                # also set current name for the current object
                self[key].name = item.name

            else:  # new item
                self[key] = item

        def removeEmptyBranches(self):
            toRemove = []
            for child in self.values():
                if not child.isLeaf():
                    child.removeEmptyBranches()
                    if list(child.values()) == []:
                        toRemove.append(child)
            for item in toRemove:
                del self[item.id]

        def sort(self, key=None, reverse=False):
            """Recursive sort of the tree: items are sorted in all branches.
            """
            ObservableSortedDictionary.sort(self, self._keyItems)
            for item in self.values():
                if not item.isLeaf():
                    item.sort(key=key, reverse=reverse)

        def _keyItems(self, id):
            '''Sorting key function'''
            i1 = self[id]
            # names are translated in lowercase to make an alphabetical sort independant of the case
            # by default (uppercase letters are < to lowecase letters)
            n1 = i1.name.lower()
            # print "comp", i1.name,i1.isLeaf(), i2.name,i2.isLeaf(), res
            # leafs must go after subtrees
            return (not i1.isLeaf(), n1)


        def compItems(self, id1, id2):
            """Comparison function

            Returns

            - 1 if i1 > i2
            - -1 if i1 < i2
            - 0 if they are equal"""
            res = 0
            i1 = self[id1]
            i2 = self[id2]
            if (not i1.isLeaf() and i2.isLeaf()):
                # if one item is a branch and the other is a leaf, the branch
                # must be before the leaf
                res = -1
            elif (i1.isLeaf() and not i2.isLeaf()):
                res = 1
            else:
                # names are translated in lowercase to make an alphabetical sort independant of the case
                # by default (uppercase letters are < to lowecase letters)
                n1 = i1.name.lower()
                n2 = i2.name.lower()
                if n1 < n2:
                    res = -1
                elif n1 > n2:
                    res = 1
            # print "comp", i1.name,i1.isLeaf(), i2.name,i2.isLeaf(), res
            return res

        def addListenerRec(self, listener):
            self.addListener(listener)
            for item in self.values():
                if not item.isLeaf():
                    item.addListenerRec(listener)

        def onAttributeChangeRec(self, attributeName, listener):
            self.onAttributeChange(attributeName, listener)
            for item in self.values():
                if not item.isLeaf():
                    item.onAttributeChangeRec(attributeName, listener)

    #-------------------------------------------------------------------------
    class Leaf(Item):

        """A tree item that cannot have children items"""

        def __init__(self, name="new", id=None, icon=None, tooltip=None, copyEnabled=True, modifiable=True, delEnabled=True, visible=True, enabled=True):
            super(EditableTree.Leaf, self).__init__(
                name, id, icon, tooltip, copyEnabled, modifiable, delEnabled, visible, enabled)

        def isLeaf(self):
            return True

    #-------------------------------------------------------------------------


class ObservableNotifier(Notifier):

    """
    This notifier can notify when the first listener is added and when the last
    listener is removed.

    It enables to use the notifier only when there are some listeners
    registred.

    Attributes
    ----------
    onAddFirstListener: :class:`Notifier`
        register a listener on this notifier to be called when the first
        listener is registred on ObservableNotifier
    onRemoveLastListener: :class:`Notifier`
        register a listener on this notifier to be called when the last
        listener is removed from ObservableNotifier
    """

    def __init__(self, parameterCount=None):
        """
        Parameters
        ----------
        parameterCount: int
            if not *None*, each registered function must be callable with that
            number of arguments (checking is done on registration).
        """
        Notifier.__init__(self, parameterCount)
        self.onAddFirstListener = Notifier()
        self.onRemoveLastListener = Notifier()

    def add(self, listener):
        nbListenersBefore = len(self._listeners)
        Notifier.add(self, listener)
        if nbListenersBefore == 0:  # before add : 0 listener, after : 1 listener -> add first listener
            if len(self._listeners) == 1:
                self.onAddFirstListener.notify()

    def remove(self, listener):
        nbListenersBefore = len(self._listeners)
        Notifier.remove(self, listener)
        if nbListenersBefore == 1:  # before add : 1 listener, after : 0 listener -> remove last listener
            if len(self._listeners) == 0:
                self.onRemoveLastListener.notify()



# -*- coding: utf-8 -*-

#  This software and supporting documentation are distributed by
#      Institut Federatif de Recherche 49
#      CEA/NeuroSpin, Batiment 145,
#      91191 Gif-sur-Yvette cedex
#      France
#
# This software is governed by the CeCILL-B license under
# French law and abiding by the rules of distribution of free software.
# You can  use, modify and/or redistribute the software under the
# terms of the CeCILL-B license as circulated by CEA, CNRS
# and INRIA at the following URL "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL-B license and that you accept its terms.

'''
This module contains text related functions.

* author: Yann Cointepas
* organization: NeuroSpin
* license: `CeCILL B <http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html>`_
'''
__docformat__ = "restructuredtext en"


def string_to_list(s):
    """
    """
    s = s.strip(' ')
    while s:
        if s[0] == '"':
            i = 1
            while True:
                j = s[i:].find('"')
                if j == -1:
                    # Syntax error, missing closing quote.
                    # return the rest of the string and exit
                    yield s[1:]
                    return
                if s[i + j - 1] == '\\':
                    i = i + j + 1
                else:
                    break
            yield s[1: i + j]
            s = s[i + j + 2:].strip(' ')
        else:
            l = s.split(' ', 1)
            yield l[0]
            if len(l) > 1:
                s = l[1].strip(' ')
            else:
                s = ''


def list_to_string(l):
    """
    """
    return ' '.join((quote_string(i) for i in l))


def quote_string(unquoted):
    """
    """
    if unquoted:
        if unquoted.find(' ') > -1:
            return '"' + unquoted.replace('"', '\\"') + '"'
        else:
            return unquoted
    return ''


def unquote_string(quoted):
    """
    """
    if quoted and quoted[0] == '"':
        if quoted[-1] == '"':
            quoted = quoted[1: -1]
        else:
            # There is a syntax error here, quoted should
            # be terminated by a double quote
            quoted = quoted[1:]
        return eval('r"' + quoted + '"').replace('\\"', '"')
    else:
        return quoted



# -*- coding: utf-8 -*-

#  This software and supporting documentation are distributed by
#      Institut Federatif de Recherche 49
#      CEA/NeuroSpin, Batiment 145,
#      91191 Gif-sur-Yvette cedex
#      France
#
# This software is governed by the CeCILL-B license under
# French law and abiding by the rules of distribution of free software.
# You can  use, modify and/or redistribute the software under the
# terms of the CeCILL-B license as circulated by CEA, CNRS
# and INRIA at the following URL "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL-B license and that you accept its terms.

'''
Reading of XML minf format.

* author: Yann Cointepas
* organization: `NeuroSpin <http://www.neurospin.org>`_
* license: `CeCILL B <http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html>`_
'''
from __future__ import absolute_import
import six
__docformat__ = "restructuredtext en"

import sys
import os
import operator
import codecs
import types
import gzip
from xml.sax.saxutils import quoteattr as xml_quoteattr
from xml.sax.saxutils import escape as xml_escape
from xml.sax import make_parser
from xml.sax.handler import ContentHandler
from xml.sax.handler import ErrorHandler

from soma.translation import translate as _
from soma.undefined import Undefined
from soma.minf.error import MinfError

from soma.minf.reader import MinfReader
from soma.minf.tree import minfStructure, listStructure, dictStructure, \
    StartStructure, EndStructure, Reference
from soma.minf.xhtml import XHTML

# This module only contains a definition of XML tags and attributes.
# It is designed to allow "import *".
from soma.minf.xml_tags import *
from io import StringIO

#------------------------------------------------------------------------------


class MinfXMLReader(MinfReader, ErrorHandler):

    '''
    Specialization of L{MinfReader} class for reading XML minf format.
    '''
    name = 'XML'

    class SaxHandler(ContentHandler):

        def __init__(self, parser):
            self.parser = parser
            self.parser._handler = MinfXMLHandler(parser)
            self.parser._stack = []

        def startElement(self, name, attrs):
            # Convert to builtin Python types
            name = six.text_type(name)
            attrs = dict([(six.text_type(i[0]), six.text_type(i[1]))
                         for i in attrs.items()])
            self.parser._stack.append(name)
            self.parser._handler.startElement(self.parser, name, attrs)

        def endElement(self, name):
            self.parser._handler.endElement(self.parser, six.text_type(name))
            self.parser._stack.pop()

        def characters(self, content):
            self.parser._handler.characters(self.parser, six.text_type(content))

    def reduction(self, source):
        self._nodesToProduce = []
        self._sax_parser = make_parser()
        self._sax_parser.setContentHandler(MinfXMLReader.SaxHandler(self))
        self._sax_parser.setErrorHandler(self)
        self._sax_parser.reset()
        self._minfStarted = False
        self._minfFinished = False
        self._nodeIdentifier = 0
        buffer = []
        while not self._nodesToProduce:
            line = source.readline()
            buffer.append(line)
            if not line:
                # end of file
                break
            self._sax_parser.feed(line)
        if self._nodesToProduce:
            return (self._nodesToProduce[0].attributes['reduction'], ''.join(buffer))
        return (None, ''.join(buffer))

    def nodeIterator(self, source):
        self._nodesToProduce = []
        self._sax_parser = make_parser()
        self._sax_parser.setContentHandler(MinfXMLReader.SaxHandler(self))
        self._sax_parser.setErrorHandler(self)
        self._sax_parser.reset()
        # self._sax_parser.feed( encodingLine )
        self._minfStarted = False
        self._minfFinished = False
        self._nodeIdentifier = 0
        while not self._minfFinished:
            while self._nodesToProduce:
                node = self._nodesToProduce.pop(0)
                yield node
            line = source.readline()
            if not line:
                # end of file
                if not self._minfStarted:
                    # Log files may be read while not finished, therefore the closing
                    # tag is missing. It is append to avoid XML parser error
                    # message.
                    self._sax_parser.feed('<' + minfTag + '/>')
                self._minfFinished = True
                break
            self._sax_parser.feed(line)
        while self._nodesToProduce:
            node = self._nodesToProduce.pop(0)
            yield node

    def parseError(self, errorMessage):
        self.fatalError(errorMessage)

    def fatalError(self, error):
        raise MinfError(_('XML parse error: %s') %
                       (six.text_type(error), ) + ' (stack = ' +
                        ','.join(['"' + i + '"' for i in self._stack]) + ')')

    def checkNoMoreAttributes(self, attributes):
        if attributes:
            self.parseError(_('Invalid attribute": %(attributes)s') %
                            {'attributes': ', '.join(['"' + n + '"' for n in attributes])})


#------------------------------------------------------------------------------
class XMLHandler(object):

    def __init__(self, parser, parent, name, attributes):
        self.parent = parent
        parser.checkNoMoreAttributes(attributes)

    def startElement(self, parser, name, attributes):
        parser.parseError(_('Unexpected tag "%s"') % (name, ))

    def endElement(self, parser, name):
        parser._handler = self.parent
        self.parent = None

    def characters(self, parser, content):
        pass


#------------------------------------------------------------------------------
class NoneXMLHandler(XMLHandler):

    def endElement(self, parser, name):
        parser._nodesToProduce.append(None)
        XMLHandler.endElement(self, parser, name)


#------------------------------------------------------------------------------
class TrueXMLHandler(XMLHandler):

    def endElement(self, parser, name):
        parser._nodesToProduce.append(True)
        XMLHandler.endElement(self, parser, name)


#------------------------------------------------------------------------------
class FalseXMLHandler(XMLHandler):

    def endElement(self, parser, name):
        parser._nodesToProduce.append(False)
        XMLHandler.endElement(self, parser, name)


#------------------------------------------------------------------------------
class NumberXMLHandler(XMLHandler):

    def __init__(self, parser, parent, name, attributes):
        XMLHandler.__init__(self, parser, parent, name, attributes)
        self._stringValue = []

    def characters(self, parser, content):
        self._stringValue.append(content)

    def endElement(self, parser, name):
        stringValue = ''.join(self._stringValue)
        ttypes = six.integer_types + (float,)
        for ttype in ttypes:
            try:
                value = ttype(stringValue)
                break
            except ValueError:
                pass
        else:
            raise
        parser._nodesToProduce.append(value)
        XMLHandler.endElement(self, parser, name)


#------------------------------------------------------------------------------
class StringXMLHandler(XMLHandler):

    def __init__(self, parser, parent, name, attributes):
        XMLHandler.__init__(self, parser, parent, name, attributes)
        self._value = []

    def characters(self, parser, content):
        self._value.append(content)

    def endElement(self, parser, name):
        parser._nodesToProduce.append(''.join(self._value))
        XMLHandler.endElement(self, parser, name)


#------------------------------------------------------------------------------
class ListXMLHandler(XMLHandler):

    def __init__(self, parser, parent, name, attributes):
        length = attributes.pop(lengthAttribute, None)
        identifier = attributes.pop(identifierAttribute, None)
        if length is not None:
            parser._nodesToProduce.append(StartStructure(listStructure,
                                                         identifier=identifier,
                                                         length=length))
        else:
            parser._nodesToProduce.append(StartStructure(listStructure,
                                                         identifier=identifier))
        XMLHandler.__init__(self, parser, parent, name, attributes)

    def startElement(self, parser, name, attributes):
        newHandler = MinfXMLHandler.getHandler(parser, self, name, attributes)
        if newHandler is None:
            parser.parseError(_('Unexpected tag "%s"') % (name, ))
        parser._handler = newHandler

    def endElement(self, parser, name):
        parser._nodesToProduce.append(EndStructure(listStructure))
        XMLHandler.endElement(self, parser, name)


#------------------------------------------------------------------------------
class DictionaryXMLHandler(XMLHandler):

    def __init__(self, parser, parent, name, attributes):
        length = attributes.pop(lengthAttribute, None)
        identifier = attributes.pop(identifierAttribute, None)
        if length is not None:
            parser._nodesToProduce.append(StartStructure(dictStructure,
                                                         identifier=identifier,
                                                         length=length))
        else:
            parser._nodesToProduce.append(StartStructure(dictStructure,
                                                         identifier=identifier))
        XMLHandler.__init__(self, parser, parent, name, attributes)

    def startElement(self, parser, name, attributes):
        nameAttr = attributes.pop(nameAttribute, None)
        if nameAttr is not None:
            parser._nodesToProduce.append(nameAttr)
        newHandler = MinfXMLHandler.getHandler(parser, self, name, attributes)
        if newHandler is None:
            parser.parseError(_('Unexpected tag "%s"') % (name, ))
        parser._handler = newHandler

    def endElement(self, parser, name):
        parser._nodesToProduce.append(EndStructure(dictStructure))
        XMLHandler.endElement(self, parser, name)


#------------------------------------------------------------------------------
class FactoryXMLHandler(XMLHandler):

    def __init__(self, parser, parent, name, attributes):
        self.type = attributes.pop(objectTypeAttribute, None)
        if self.type is None:
            parser.parseError(
                _('%s attribute missing') % (objectTypeAttribute, ))
        identifier = attributes.pop(identifierAttribute, None)
        parser._nodesToProduce.append(
            StartStructure(self.type, identifier=identifier))
        XMLHandler.__init__(self, parser, parent, name, attributes)

    def startElement(self, parser, name, attributes):
        nameAttr = attributes.pop(nameAttribute, None)
        parser._nodesToProduce.append(nameAttr)
        newHandler = MinfXMLHandler.getHandler(parser, self, name, attributes)
        if newHandler is None:
            parser.parseError(_('Unexpected tag "%s"') % (name, ))
        parser._handler = newHandler

    def endElement(self, parser, name):
        parser._nodesToProduce.append(EndStructure(self.type))
        XMLHandler.endElement(self, parser, name)


#------------------------------------------------------------------------------
class ReferenceXMLHandler(XMLHandler):

    def __init__(self, parser, parent, name, attributes):
        identifier = attributes.pop(identifierAttribute, None)
        if self.type is None:
            parser.parseError(
                _('%s attribute missing') % (identifierAttribute, ))
        parser._nodesToProduce.append(Reference(identifier=identifier))
        XMLHandler.__init__(self, parser, parent, name, attributes)

    def endElement(self, parser, name):
        XMLHandler.endElement(self, parser, name)


#------------------------------------------------------------------------------
class XHTMLHandler(XMLHandler):

    def __init__(self, parser, parent, name, attributes):
        self.stack = [XHTML(name, attributes)]
        self._characters = []
        XMLHandler.__init__(self, parser, parent, name, attributes)

    def characters(self, parser, content):
        self._characters.append(content)

    def startElement(self, parser, name, attributes):
        c = ''.join(self._characters)
        if c:
            self.stack[-1].content.append(c)
        self._characters = []
        newItem = XHTML(name, attributes)
        self.stack[-1].content.append(newItem)
        self.stack.append(newItem)

    def endElement(self, parser, name):
        item = self.stack.pop()
        c = ''.join(self._characters)
        self._characters = []
        if c:
            item.content.append(c)
        if not self.stack:
            parser._nodesToProduce.append(item)
            XMLHandler.endElement(self, parser, name)


#------------------------------------------------------------------------------
class MinfXMLHandler(XMLHandler):
    _handlers = {
        noneTag: NoneXMLHandler,
        trueTag: TrueXMLHandler,
        falseTag: FalseXMLHandler,
        numberTag: NumberXMLHandler,
        stringTag: StringXMLHandler,
        listTag: ListXMLHandler,
        dictionaryTag: DictionaryXMLHandler,
        factoryTag: FactoryXMLHandler,
        referenceTag: ReferenceXMLHandler,
        xhtmlTag: XHTMLHandler,
    }

    def __init__(self, parser):
        XMLHandler.__init__(self, parser, None, minfTag, {})

    def getHandler(parser, parent, name, attributes):
        handler = MinfXMLHandler._handlers.get(name)
        if handler is not None:
            handler = handler(parser, parent, name, attributes)
        return handler
    getHandler = staticmethod(getHandler)

    def startElement(self, parser, name, attributes):
        if parser._minfStarted:
            nameAttr = attributes.pop(nameAttribute, None)
            if nameAttr is None:
                if self._obsoleteFormat:
                    parser.parseError(
                        _('%s attribute required for minf_1.0') % (nameAttribute,))
            else:
                if self._obsoleteFormat:
                    parser._nodesToProduce.append(nameAttr)
                else:
                    parser.parseError(
                        _('Unexpected attribute %s') % (nameAttribute, ))
            newHandler = MinfXMLHandler.getHandler(
                parser, self, name, attributes)
            if newHandler is None:
                parser.parseError(_('Unexpected tag "%s"') % (name, ))
            parser._handler = newHandler
        else:
            if name != minfTag:
                # Document is not a minf file
                parser.parseError(
                    _('Wrong document type, expecting "%(minf)s" instead of '
                        '"%(other)s>"') % {'minf': minfTag, 'other': name})
            # Checking minf format
            self._obsoleteFormat = False
            expanderName = attributes.pop(expanderAttribute, None)
            if expanderName is None:
                # Compatibility with obsolete minf 1.0 XML format
                version = attributes.pop('version', None)
                if version is None:
                    expanderName = 'minf_2.0'
                else:
                    if version != '1.0':
                        parser.parseError(_('Wrong value for attribute "version", found '
                                            '"%s" but only "1.0" is accepted') %
                                          (version, ))
                    self._obsoleteFormat = True
                    expanderName = 'minf_1.0'
            parser._nodesToProduce.append(
                StartStructure(minfStructure, reduction=expanderName))
            # Compatibility with obsolete minf 1.0 XML format
            if self._obsoleteFormat:
                parser._nodesToProduce.append(StartStructure(dictStructure))
            parser._minfStarted = True
            # Check attributes
            parser.checkNoMoreAttributes(attributes)

    def endElement(self, parser, name):
        if name == minfTag:
            # Compatibility with obsolete minf 1.0 XML format
            if self._obsoleteFormat:
                parser._nodesToProduce.append(EndStructure(dictStructure))
            parser._nodesToProduce.append(EndStructure(minfStructure))
            parser._minfFinished = True
        else:
            parser._nodesToProduce.append(EndStructure(minfStructure))



# -*- coding: utf-8 -*-

from __future__ import print_function

from __future__ import absolute_import
import unittest
import shutil
import os
import tempfile
from soma import application
from soma import fom
import sys


class TestFOM(unittest.TestCase):

    def setUp(self):
        self.work_dir = tempfile.mkdtemp(prefix='soma_test_fom')

    def tearDown(self):
        try:
            shutil.rmtree(self.work_dir)
            #pass
        except OSError:
            pass

    def test_fom(self):
        app = application.Application('testapp', plugin_modules=['soma.fom'])
        if 'soma.fom' not in app.loaded_plugin_modules:
            app.initialize()
        app.fom_path = [os.path.join(self.work_dir, 'foms')]
        app.fom_manager.paths = app.fom_path # BUG: should be automatic
        os.mkdir(app.fom_path[0])
        #print('fom_path:', app.fom_path, file=sys.stderr)
        fom_filename = 'test_fom'
        open(os.path.join(app.fom_path[0], fom_filename + '.json'), 'w').write(
            '''{
    "fom_name": "test_fom",

    "formats": {
        "NIFTI": "nii",
        "GIS": "ima"
    },

    "format_lists": {
        "images": ["NIFTI", "GIS"]
    },

    "attribute_definitions": {
        "acquisition" : {"default_value" : "default_acquisition"},
        "analysis" : {"default_value" : "default_analysis"},
        "sulci_recognition_session" :  {"default_value" : "default_session"},
        "graph_version": {"default_value": "3.1"}
    },

    "shared_patterns": {
      "acquisition": "<center>/<subject>/t1mri/<acquisition>",
      "analysis": "{acquisition}/<analysis>",
      "recognition_analysis": "{analysis}/folds/<graph_version>/<sulci_recognition_session>_auto"
    },

    "processes": {
        "Morphologist": {
            "t1mri":
                [["input:{acquisition}/<subject>", "images"]]
        }
    }
}
'''
        )
        foms = app.fom_manager.load_foms(fom_filename)
        atp = fom.AttributesToPaths(foms)
        pta = fom.PathToAttributes(foms)


def test():
    suite = unittest.TestLoader().loadTestsFromTestCase(TestFOM)
    runtime = unittest.TextTestRunner(verbosity=2).run(suite)
    return runtime.wasSuccessful()


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function

from soma.qt_gui.qt_backend import Qt
import six
try:
    from traits.api import Undefined
except ImportError:
    Undefined = None
from functools import partial


class QPredefLineEdit(Qt.QLineEdit):
    '''
    A QLineEdit variant allowing to select predefined values using a right-
    click popup.

    This allows to handle and distinguish between values such as "", None, and
    traits.Undefined typically, or allow other values.

    The widget also supports None and Undefined values, represented as "<none>"
    and "<undefined>" respectively, when allow_none and/or allow_undefined are
    enabled. These special values can be queried using :meth:`value` instead of
    :meth:`text`.
    '''
    def __init__(self, parent_or_contents=None, parent=None,
                 predefined_values=None, allow_none=False,
                 allow_undefined=False, *args, **kwargs):
        if parent is not None:
            args = (parent, ) + args
        if parent_or_contents is not None:
            args = (parent_or_contents, ) + args
        super(QPredefLineEdit, self).__init__(*args, **kwargs)
        if predefined_values is None:
            predefined_values = []
        self.predefined_values = predefined_values
        self.allow_none = allow_none or None in predefined_values \
            or ('<none>', None) in predefined_values
        self.allow_undefined = allow_undefined \
            or Undefined in predefined_values \
            or ('<undefined>', Undefined) in predefined_values

    def contextMenuEvent(self, event):
        menu = self.createStandardContextMenu()

        if self.predefined_values:
            menu.addSeparator()
            for value in self.predefined_values:
                if isinstance(value, (tuple, list)):
                    msg, value = value
                else:
                    msg = value
                msg = self.text_repr(msg)
                action = menu.addAction(msg)
                action.value = value
                action.triggered.connect(partial(self.set_value, value))
        menu.exec_(event.globalPos())

    def text_repr(self, value):
        if value is None:
            return '<none>'
        if value is Undefined:
            return '<undefined>'
        if not isinstance(value, (six.string_types, type(b''))):
            value = str(value)
        return six.ensure_text(value)

    def value(self):
        text = self.text()
        if self.allow_none and text == '<none>':
            return None
        if self.allow_undefined and text == '<undefined>':
            return Undefined
        return text

    def set_value(self, value):
        self.setText(self.text_repr(value))



# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import absolute_import
import os, sys
import logging


def load_pilots(root, path):
    """ Load all the modules in the use_cases module: path is
    recursively scanned for __init__.py files.
    Any function declared inside will be loaded.

    Parameters
    ----------
    root : str (mandatory)
        path to the use_cases module.
    path : str
        path to the module

    Returns
    -------
    pilots : dict
        a dict with module name as keys referencing to function module used
        for unitest.
    """

    pilots = {}
    files = os.listdir(path)

    for fname in files:
        if os.path.isdir(os.path.join(path, fname)):
            sub_pilots = load_pilots(root, os.path.join(path, fname))
            pilots.update(sub_pilots)

    if not any([x in files for x in ["__init__.py", ]]):
        # No __init__ file
        return pilots

    for fname in files:
        if fname.endswith(".py") and fname.startswith("test_"):

            module_name = (["soma"] +
                path[len(os.path.normpath(root)) + 1:].split(os.path.sep) +
                [os.path.splitext(fname)[0]])
            module_name = ".".join([x for x in module_name if x])

            try:
                __import__(module_name)
            except ImportError as e:
                # An api exists, but it cannot be imported
                logging.debug("Could not import {0}:"
                              "{1}".format(module_name, e))
                continue

            module = sys.modules[module_name]

            for function in dir(module):
                if function in ["test", ]:
                    if module_name in  list(pilots.keys()):
                        pilots[module_name].append(getattr(module, function))
                    else:
                        pilots[module_name] = [getattr(module, function), ]

    return pilots


if __name__ == "__main__":

    import soma
    module_path = soma.__path__[0]
    print(load_pilots(module_path, module_path))



# -*- coding: utf-8 -*-

#  This software and supporting documentation are distributed by
#      Institut Federatif de Recherche 49
#      CEA/NeuroSpin, Batiment 145,
#      91191 Gif-sur-Yvette cedex
#      France
#
# This software is governed by the CeCILL-B license under
# French law and abiding by the rules of distribution of free software.
# You can  use, modify and/or redistribute the software under the
# terms of the CeCILL-B license as circulated by CEA, CNRS
# and INRIA at the following URL "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL-B license and that you accept its terms.

'''
Writing of XML minf format.

* author: Yann Cointepas
* organization: `NeuroSpin <http://www.neurospin.org>`_
* license: `CeCILL B <http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html>`_
'''
from __future__ import absolute_import
__docformat__ = "restructuredtext en"

import codecs
import six
from six.moves import range
from xml.sax.saxutils import quoteattr as xml_quoteattr
from xml.sax.saxutils import escape as xml_escape
from soma.translation import translate as _
from soma.minf.tree import createMinfReducer
from soma.minf.writer import MinfWriter
from soma.minf.tree import minfStructure, listStructure, dictStructure, \
    StartStructure, EndStructure
from soma.minf.error import MinfError
from soma.undefined import Undefined
import sys


# This module only contains a definition of XML tags and attributes.
# It is designed to allow "import *".
from soma.minf.xml_tags import *

#: Replacement table for characters that are not allowed in XML
xml_replacement = dict([(eval('"\\x' + ('0' + hex(i)[2:])[-2:] + '"'), '')
                       for i in range(32)])
del xml_replacement['\x09']
del xml_replacement['\x0a']
del xml_replacement['\x0d']


#------------------------------------------------------------------------------
class MinfXMLWriter(MinfWriter):

    '''
    Specialization of L{MinfWriter} class for writing XML minf format.
    '''

    name = 'XML'

    def __init__(self, file, reducer,
                 encoding='utf-8',
                 level=0,
                 append=False):
        self.__file = file
        self.reducer = createMinfReducer(reducer)
        self.encoder = codecs.getencoder(encoding)
        self.level = level
        self.indentString = '  '
        if not append:
            self._writeLine('<?xml version="1.0" encoding=' +
                            xml_quoteattr(encoding) + ' ?>')
            self._encodeAndWriteLine('<' + minfTag + ' ' + expanderAttribute +
                                     '=' + xml_quoteattr(reducer) + '>')

    def close(self):
        """Close the Minf sytax tree. The underlying file is NOT closed."""
        if self.__file is not None:
            self.__file.flush()
            self._encodeAndWriteLine('</' + minfTag + '>')
            self.__file = None

    def write(self, value):
        minfNodeIterator = self.reducer.reduce(value)
        for minfNode in minfNodeIterator:
            self._write(minfNodeIterator, minfNode, 0, None)

    def _write(self, minfNodeIterator, minfNode, level, name):
        if minfNode is Undefined:
            minfNode = next(minfNodeIterator)
        attributes = {}
        if name is not None:
            attributes[nameAttribute] = name
        if isinstance(minfNode, StartStructure):
            if minfNode.type == listStructure:
                naming = False
                stringNaming = False
                length = minfNode.attributes.get('length')
                if length:
                    attributes[lengthAttribute] = length
                tag = listTag
            elif minfNode.type == dictStructure:
                naming = True
                stringNaming = False
                length = minfNode.attributes.get('length')
                if length:
                    attributes[lengthAttribute] = length
                tag = dictionaryTag
            else:
                naming = True
                stringNaming = True
                tag = factoryTag
                attributes[objectTypeAttribute] = minfNode.type
            if attributes:
                attributes = ' ' + \
                    ' '.join([n + '=' + xml_quoteattr(six.text_type(v))
                             for n, v in six.iteritems(attributes)])
            else:
                attributes = ''
            self._encodeAndWriteLine('<' + tag + attributes + '>', level)
            ntype = minfNode.type
            for minfNode in minfNodeIterator:
                if isinstance(minfNode, EndStructure):
                    if ntype != minfNode.type:
                        raise MinfError(_('Wrong Minf structure ending, expecting %(exp)s instead of %(rcv)s') %
                                        {'exp': ntype, 'rcv': minfNode.type})
                    self._encodeAndWriteLine('</' + tag + '>', level)
                    break
                elif naming:
                    if isinstance(minfNode, six.string_types):
                        self._write(
                            minfNodeIterator, Undefined, level + 1, minfNode)
                    elif minfNode is None:
                        if not stringNaming:
                            self._write(
                                minfNodeIterator, minfNode, level + 1, None)
                        self._write(
                            minfNodeIterator, Undefined, level + 1, None)
                    else:
                        self._write(
                            minfNodeIterator, minfNode, level + 1, None)
                        self._write(
                            minfNodeIterator, Undefined, level + 1, None)
                else:
                    self._write(minfNodeIterator, minfNode, level + 1, None)
        elif isinstance(minfNode, EndStructure):
            raise MinfError(
                _('Unexpected Minf structure ending: %s') % (minfNode.type, ))
            level -= 1
        else:
            if attributes:
                attributesXML = ' ' + \
                    ' '.join([n + '=' + xml_quoteattr(six.text_type(v))
                             for n, v in six.iteritems(attributes)])
            else:
                attributesXML = ''
            if minfNode is None:
                self._encodeAndWriteLine(
                    '<' + noneTag + attributesXML + '/>', level)
            elif isinstance(minfNode, bool):
                if minfNode:
                    self._encodeAndWriteLine(
                        '<' + trueTag + attributesXML + '/>', level)
                else:
                    self._encodeAndWriteLine(
                        '<' + falseTag + attributesXML + '/>', level)
            elif isinstance(minfNode, (float,) + six.integer_types):
                self._encodeAndWriteLine('<' + numberTag + attributesXML + '>' + six.text_type(minfNode) + '</' +
                                         numberTag + '>', level)
            elif isinstance(minfNode, six.string_types):

                if type(minfNode) is six.binary_type:
                    try:
                        minfNode = minfNode.decode("utf-8")
                    except UnicodeDecodeError:
                        minfNode = minfNode.decode("iso-8859-1")
                self._encodeAndWriteLine('<' + stringTag + attributesXML + '>' +
                                         xml_escape(minfNode, xml_replacement) + '</' + stringTag + '>', level)
            elif hasattr(minfNode, '__minfxml__'):
                minfNode.__minfxml__(self, attributes, level)
            else:
                raise MinfError(
                    _('Cannot save an object of type %s as an XML atom') % (str(type(minfNode)), ))

    def _encodeAndWriteLine(self, line, level=0):
        self._writeLine(self.encoder(line)[0], level=level)

    def _writeLine(self, line, level=0):
        if self.level is None:
            indent = ''
            nl = ''
            level = None
        else:
            indent = self.indentString * (self.level + level)
            nl = '\n'
        line = six.ensure_text(line, 'utf8')
        try:
            self.__file.write(indent + line + nl)
        except TypeError:
            # in python3 writing in a binary stream needs to write byte
            # objects, not strings.
            # however there is no [obvious] way to know if the file object
            # is open in string or binary mode, and thus what it expects.
            # if you want my opinion, it's completely crazy...
            self.__file.write((indent + line + nl).encode())

    def flush(self):
        self.__file.flush()

    def change_file(self, file):
        self.__file = file
